// Generated by scripts/grapheme.py — do not edit.
// Unicode 17.0.0: GraphemeBreakProperty + Extended_Pictographic + InCB.

use super::grapheme::GraphemeClass;

/// The Unicode version of the grapheme-break data; always equal
/// to the width tables' [`UNICODE_VERSION`](crate::UNICODE_VERSION).
pub const GRAPHEME_UNICODE_VERSION: (u8, u8, u8) = (17, 0, 0);

/// Sorted, non-overlapping ranges; codepoints not listed are
/// `GraphemeClass::Other`.
pub(crate) static GRAPHEME_CLASS_RANGES: &[(u32, u32, GraphemeClass)] = &[
    (0x0, 0x9, GraphemeClass::Control),
    (0xA, 0xA, GraphemeClass::LF),
    (0xB, 0xC, GraphemeClass::Control),
    (0xD, 0xD, GraphemeClass::CR),
    (0xE, 0x1F, GraphemeClass::Control),
    (0x7F, 0x9F, GraphemeClass::Control),
    (0xA9, 0xA9, GraphemeClass::ExtPic),
    (0xAD, 0xAD, GraphemeClass::Control),
    (0xAE, 0xAE, GraphemeClass::ExtPic),
    (0x300, 0x36F, GraphemeClass::ExtendIncb),
    (0x483, 0x489, GraphemeClass::ExtendIncb),
    (0x591, 0x5BD, GraphemeClass::ExtendIncb),
    (0x5BF, 0x5BF, GraphemeClass::ExtendIncb),
    (0x5C1, 0x5C2, GraphemeClass::ExtendIncb),
    (0x5C4, 0x5C5, GraphemeClass::ExtendIncb),
    (0x5C7, 0x5C7, GraphemeClass::ExtendIncb),
    (0x600, 0x605, GraphemeClass::Prepend),
    (0x610, 0x61A, GraphemeClass::ExtendIncb),
    (0x61C, 0x61C, GraphemeClass::Control),
    (0x64B, 0x65F, GraphemeClass::ExtendIncb),
    (0x670, 0x670, GraphemeClass::ExtendIncb),
    (0x6D6, 0x6DC, GraphemeClass::ExtendIncb),
    (0x6DD, 0x6DD, GraphemeClass::Prepend),
    (0x6DF, 0x6E4, GraphemeClass::ExtendIncb),
    (0x6E7, 0x6E8, GraphemeClass::ExtendIncb),
    (0x6EA, 0x6ED, GraphemeClass::ExtendIncb),
    (0x70F, 0x70F, GraphemeClass::Prepend),
    (0x711, 0x711, GraphemeClass::ExtendIncb),
    (0x730, 0x74A, GraphemeClass::ExtendIncb),
    (0x7A6, 0x7B0, GraphemeClass::ExtendIncb),
    (0x7EB, 0x7F3, GraphemeClass::ExtendIncb),
    (0x7FD, 0x7FD, GraphemeClass::ExtendIncb),
    (0x816, 0x819, GraphemeClass::ExtendIncb),
    (0x81B, 0x823, GraphemeClass::ExtendIncb),
    (0x825, 0x827, GraphemeClass::ExtendIncb),
    (0x829, 0x82D, GraphemeClass::ExtendIncb),
    (0x859, 0x85B, GraphemeClass::ExtendIncb),
    (0x890, 0x891, GraphemeClass::Prepend),
    (0x897, 0x89F, GraphemeClass::ExtendIncb),
    (0x8CA, 0x8E1, GraphemeClass::ExtendIncb),
    (0x8E2, 0x8E2, GraphemeClass::Prepend),
    (0x8E3, 0x902, GraphemeClass::ExtendIncb),
    (0x903, 0x903, GraphemeClass::SpacingMark),
    (0x915, 0x939, GraphemeClass::Consonant),
    (0x93A, 0x93A, GraphemeClass::ExtendIncb),
    (0x93B, 0x93B, GraphemeClass::SpacingMark),
    (0x93C, 0x93C, GraphemeClass::ExtendIncb),
    (0x93E, 0x940, GraphemeClass::SpacingMark),
    (0x941, 0x948, GraphemeClass::ExtendIncb),
    (0x949, 0x94C, GraphemeClass::SpacingMark),
    (0x94D, 0x94D, GraphemeClass::Linker),
    (0x94E, 0x94F, GraphemeClass::SpacingMark),
    (0x951, 0x957, GraphemeClass::ExtendIncb),
    (0x958, 0x95F, GraphemeClass::Consonant),
    (0x962, 0x963, GraphemeClass::ExtendIncb),
    (0x978, 0x97F, GraphemeClass::Consonant),
    (0x981, 0x981, GraphemeClass::ExtendIncb),
    (0x982, 0x983, GraphemeClass::SpacingMark),
    (0x995, 0x9A8, GraphemeClass::Consonant),
    (0x9AA, 0x9B0, GraphemeClass::Consonant),
    (0x9B2, 0x9B2, GraphemeClass::Consonant),
    (0x9B6, 0x9B9, GraphemeClass::Consonant),
    (0x9BC, 0x9BC, GraphemeClass::ExtendIncb),
    (0x9BE, 0x9BE, GraphemeClass::ExtendIncb),
    (0x9BF, 0x9C0, GraphemeClass::SpacingMark),
    (0x9C1, 0x9C4, GraphemeClass::ExtendIncb),
    (0x9C7, 0x9C8, GraphemeClass::SpacingMark),
    (0x9CB, 0x9CC, GraphemeClass::SpacingMark),
    (0x9CD, 0x9CD, GraphemeClass::Linker),
    (0x9D7, 0x9D7, GraphemeClass::ExtendIncb),
    (0x9DC, 0x9DD, GraphemeClass::Consonant),
    (0x9DF, 0x9DF, GraphemeClass::Consonant),
    (0x9E2, 0x9E3, GraphemeClass::ExtendIncb),
    (0x9F0, 0x9F1, GraphemeClass::Consonant),
    (0x9FE, 0x9FE, GraphemeClass::ExtendIncb),
    (0xA01, 0xA02, GraphemeClass::ExtendIncb),
    (0xA03, 0xA03, GraphemeClass::SpacingMark),
    (0xA3C, 0xA3C, GraphemeClass::ExtendIncb),
    (0xA3E, 0xA40, GraphemeClass::SpacingMark),
    (0xA41, 0xA42, GraphemeClass::ExtendIncb),
    (0xA47, 0xA48, GraphemeClass::ExtendIncb),
    (0xA4B, 0xA4D, GraphemeClass::ExtendIncb),
    (0xA51, 0xA51, GraphemeClass::ExtendIncb),
    (0xA70, 0xA71, GraphemeClass::ExtendIncb),
    (0xA75, 0xA75, GraphemeClass::ExtendIncb),
    (0xA81, 0xA82, GraphemeClass::ExtendIncb),
    (0xA83, 0xA83, GraphemeClass::SpacingMark),
    (0xA95, 0xAA8, GraphemeClass::Consonant),
    (0xAAA, 0xAB0, GraphemeClass::Consonant),
    (0xAB2, 0xAB3, GraphemeClass::Consonant),
    (0xAB5, 0xAB9, GraphemeClass::Consonant),
    (0xABC, 0xABC, GraphemeClass::ExtendIncb),
    (0xABE, 0xAC0, GraphemeClass::SpacingMark),
    (0xAC1, 0xAC5, GraphemeClass::ExtendIncb),
    (0xAC7, 0xAC8, GraphemeClass::ExtendIncb),
    (0xAC9, 0xAC9, GraphemeClass::SpacingMark),
    (0xACB, 0xACC, GraphemeClass::SpacingMark),
    (0xACD, 0xACD, GraphemeClass::Linker),
    (0xAE2, 0xAE3, GraphemeClass::ExtendIncb),
    (0xAF9, 0xAF9, GraphemeClass::Consonant),
    (0xAFA, 0xAFF, GraphemeClass::ExtendIncb),
    (0xB01, 0xB01, GraphemeClass::ExtendIncb),
    (0xB02, 0xB03, GraphemeClass::SpacingMark),
    (0xB15, 0xB28, GraphemeClass::Consonant),
    (0xB2A, 0xB30, GraphemeClass::Consonant),
    (0xB32, 0xB33, GraphemeClass::Consonant),
    (0xB35, 0xB39, GraphemeClass::Consonant),
    (0xB3C, 0xB3C, GraphemeClass::ExtendIncb),
    (0xB3E, 0xB3F, GraphemeClass::ExtendIncb),
    (0xB40, 0xB40, GraphemeClass::SpacingMark),
    (0xB41, 0xB44, GraphemeClass::ExtendIncb),
    (0xB47, 0xB48, GraphemeClass::SpacingMark),
    (0xB4B, 0xB4C, GraphemeClass::SpacingMark),
    (0xB4D, 0xB4D, GraphemeClass::Linker),
    (0xB55, 0xB57, GraphemeClass::ExtendIncb),
    (0xB5C, 0xB5D, GraphemeClass::Consonant),
    (0xB5F, 0xB5F, GraphemeClass::Consonant),
    (0xB62, 0xB63, GraphemeClass::ExtendIncb),
    (0xB71, 0xB71, GraphemeClass::Consonant),
    (0xB82, 0xB82, GraphemeClass::ExtendIncb),
    (0xBBE, 0xBBE, GraphemeClass::ExtendIncb),
    (0xBBF, 0xBBF, GraphemeClass::SpacingMark),
    (0xBC0, 0xBC0, GraphemeClass::ExtendIncb),
    (0xBC1, 0xBC2, GraphemeClass::SpacingMark),
    (0xBC6, 0xBC8, GraphemeClass::SpacingMark),
    (0xBCA, 0xBCC, GraphemeClass::SpacingMark),
    (0xBCD, 0xBCD, GraphemeClass::ExtendIncb),
    (0xBD7, 0xBD7, GraphemeClass::ExtendIncb),
    (0xC00, 0xC00, GraphemeClass::ExtendIncb),
    (0xC01, 0xC03, GraphemeClass::SpacingMark),
    (0xC04, 0xC04, GraphemeClass::ExtendIncb),
    (0xC15, 0xC28, GraphemeClass::Consonant),
    (0xC2A, 0xC39, GraphemeClass::Consonant),
    (0xC3C, 0xC3C, GraphemeClass::ExtendIncb),
    (0xC3E, 0xC40, GraphemeClass::ExtendIncb),
    (0xC41, 0xC44, GraphemeClass::SpacingMark),
    (0xC46, 0xC48, GraphemeClass::ExtendIncb),
    (0xC4A, 0xC4C, GraphemeClass::ExtendIncb),
    (0xC4D, 0xC4D, GraphemeClass::Linker),
    (0xC55, 0xC56, GraphemeClass::ExtendIncb),
    (0xC58, 0xC5A, GraphemeClass::Consonant),
    (0xC62, 0xC63, GraphemeClass::ExtendIncb),
    (0xC81, 0xC81, GraphemeClass::ExtendIncb),
    (0xC82, 0xC83, GraphemeClass::SpacingMark),
    (0xCBC, 0xCBC, GraphemeClass::ExtendIncb),
    (0xCBE, 0xCBE, GraphemeClass::SpacingMark),
    (0xCBF, 0xCC0, GraphemeClass::ExtendIncb),
    (0xCC1, 0xCC1, GraphemeClass::SpacingMark),
    (0xCC2, 0xCC2, GraphemeClass::ExtendIncb),
    (0xCC3, 0xCC4, GraphemeClass::SpacingMark),
    (0xCC6, 0xCC8, GraphemeClass::ExtendIncb),
    (0xCCA, 0xCCD, GraphemeClass::ExtendIncb),
    (0xCD5, 0xCD6, GraphemeClass::ExtendIncb),
    (0xCE2, 0xCE3, GraphemeClass::ExtendIncb),
    (0xCF3, 0xCF3, GraphemeClass::SpacingMark),
    (0xD00, 0xD01, GraphemeClass::ExtendIncb),
    (0xD02, 0xD03, GraphemeClass::SpacingMark),
    (0xD15, 0xD3A, GraphemeClass::Consonant),
    (0xD3B, 0xD3C, GraphemeClass::ExtendIncb),
    (0xD3E, 0xD3E, GraphemeClass::ExtendIncb),
    (0xD3F, 0xD40, GraphemeClass::SpacingMark),
    (0xD41, 0xD44, GraphemeClass::ExtendIncb),
    (0xD46, 0xD48, GraphemeClass::SpacingMark),
    (0xD4A, 0xD4C, GraphemeClass::SpacingMark),
    (0xD4D, 0xD4D, GraphemeClass::Linker),
    (0xD4E, 0xD4E, GraphemeClass::Prepend),
    (0xD57, 0xD57, GraphemeClass::ExtendIncb),
    (0xD62, 0xD63, GraphemeClass::ExtendIncb),
    (0xD81, 0xD81, GraphemeClass::ExtendIncb),
    (0xD82, 0xD83, GraphemeClass::SpacingMark),
    (0xDCA, 0xDCA, GraphemeClass::ExtendIncb),
    (0xDCF, 0xDCF, GraphemeClass::ExtendIncb),
    (0xDD0, 0xDD1, GraphemeClass::SpacingMark),
    (0xDD2, 0xDD4, GraphemeClass::ExtendIncb),
    (0xDD6, 0xDD6, GraphemeClass::ExtendIncb),
    (0xDD8, 0xDDE, GraphemeClass::SpacingMark),
    (0xDDF, 0xDDF, GraphemeClass::ExtendIncb),
    (0xDF2, 0xDF3, GraphemeClass::SpacingMark),
    (0xE31, 0xE31, GraphemeClass::ExtendIncb),
    (0xE33, 0xE33, GraphemeClass::SpacingMark),
    (0xE34, 0xE3A, GraphemeClass::ExtendIncb),
    (0xE47, 0xE4E, GraphemeClass::ExtendIncb),
    (0xEB1, 0xEB1, GraphemeClass::ExtendIncb),
    (0xEB3, 0xEB3, GraphemeClass::SpacingMark),
    (0xEB4, 0xEBC, GraphemeClass::ExtendIncb),
    (0xEC8, 0xECE, GraphemeClass::ExtendIncb),
    (0xF18, 0xF19, GraphemeClass::ExtendIncb),
    (0xF35, 0xF35, GraphemeClass::ExtendIncb),
    (0xF37, 0xF37, GraphemeClass::ExtendIncb),
    (0xF39, 0xF39, GraphemeClass::ExtendIncb),
    (0xF3E, 0xF3F, GraphemeClass::SpacingMark),
    (0xF71, 0xF7E, GraphemeClass::ExtendIncb),
    (0xF7F, 0xF7F, GraphemeClass::SpacingMark),
    (0xF80, 0xF84, GraphemeClass::ExtendIncb),
    (0xF86, 0xF87, GraphemeClass::ExtendIncb),
    (0xF8D, 0xF97, GraphemeClass::ExtendIncb),
    (0xF99, 0xFBC, GraphemeClass::ExtendIncb),
    (0xFC6, 0xFC6, GraphemeClass::ExtendIncb),
    (0x1000, 0x102A, GraphemeClass::Consonant),
    (0x102D, 0x1030, GraphemeClass::ExtendIncb),
    (0x1031, 0x1031, GraphemeClass::SpacingMark),
    (0x1032, 0x1037, GraphemeClass::ExtendIncb),
    (0x1039, 0x1039, GraphemeClass::Linker),
    (0x103A, 0x103A, GraphemeClass::ExtendIncb),
    (0x103B, 0x103C, GraphemeClass::SpacingMark),
    (0x103D, 0x103E, GraphemeClass::ExtendIncb),
    (0x103F, 0x103F, GraphemeClass::Consonant),
    (0x1050, 0x1055, GraphemeClass::Consonant),
    (0x1056, 0x1057, GraphemeClass::SpacingMark),
    (0x1058, 0x1059, GraphemeClass::ExtendIncb),
    (0x105A, 0x105D, GraphemeClass::Consonant),
    (0x105E, 0x1060, GraphemeClass::ExtendIncb),
    (0x1061, 0x1061, GraphemeClass::Consonant),
    (0x1065, 0x1066, GraphemeClass::Consonant),
    (0x106E, 0x1070, GraphemeClass::Consonant),
    (0x1071, 0x1074, GraphemeClass::ExtendIncb),
    (0x1075, 0x1081, GraphemeClass::Consonant),
    (0x1082, 0x1082, GraphemeClass::ExtendIncb),
    (0x1084, 0x1084, GraphemeClass::SpacingMark),
    (0x1085, 0x1086, GraphemeClass::ExtendIncb),
    (0x108D, 0x108D, GraphemeClass::ExtendIncb),
    (0x108E, 0x108E, GraphemeClass::Consonant),
    (0x109D, 0x109D, GraphemeClass::ExtendIncb),
    (0x1100, 0x115F, GraphemeClass::L),
    (0x1160, 0x11A7, GraphemeClass::V),
    (0x11A8, 0x11FF, GraphemeClass::T),
    (0x135D, 0x135F, GraphemeClass::ExtendIncb),
    (0x1712, 0x1715, GraphemeClass::ExtendIncb),
    (0x1732, 0x1734, GraphemeClass::ExtendIncb),
    (0x1752, 0x1753, GraphemeClass::ExtendIncb),
    (0x1772, 0x1773, GraphemeClass::ExtendIncb),
    (0x1780, 0x17B3, GraphemeClass::Consonant),
    (0x17B4, 0x17B5, GraphemeClass::ExtendIncb),
    (0x17B6, 0x17B6, GraphemeClass::SpacingMark),
    (0x17B7, 0x17BD, GraphemeClass::ExtendIncb),
    (0x17BE, 0x17C5, GraphemeClass::SpacingMark),
    (0x17C6, 0x17C6, GraphemeClass::ExtendIncb),
    (0x17C7, 0x17C8, GraphemeClass::SpacingMark),
    (0x17C9, 0x17D1, GraphemeClass::ExtendIncb),
    (0x17D2, 0x17D2, GraphemeClass::Linker),
    (0x17D3, 0x17D3, GraphemeClass::ExtendIncb),
    (0x17DD, 0x17DD, GraphemeClass::ExtendIncb),
    (0x180B, 0x180D, GraphemeClass::ExtendIncb),
    (0x180E, 0x180E, GraphemeClass::Control),
    (0x180F, 0x180F, GraphemeClass::ExtendIncb),
    (0x1885, 0x1886, GraphemeClass::ExtendIncb),
    (0x18A9, 0x18A9, GraphemeClass::ExtendIncb),
    (0x1920, 0x1922, GraphemeClass::ExtendIncb),
    (0x1923, 0x1926, GraphemeClass::SpacingMark),
    (0x1927, 0x1928, GraphemeClass::ExtendIncb),
    (0x1929, 0x192B, GraphemeClass::SpacingMark),
    (0x1930, 0x1931, GraphemeClass::SpacingMark),
    (0x1932, 0x1932, GraphemeClass::ExtendIncb),
    (0x1933, 0x1938, GraphemeClass::SpacingMark),
    (0x1939, 0x193B, GraphemeClass::ExtendIncb),
    (0x1A17, 0x1A18, GraphemeClass::ExtendIncb),
    (0x1A19, 0x1A1A, GraphemeClass::SpacingMark),
    (0x1A1B, 0x1A1B, GraphemeClass::ExtendIncb),
    (0x1A20, 0x1A54, GraphemeClass::Consonant),
    (0x1A55, 0x1A55, GraphemeClass::SpacingMark),
    (0x1A56, 0x1A56, GraphemeClass::ExtendIncb),
    (0x1A57, 0x1A57, GraphemeClass::SpacingMark),
    (0x1A58, 0x1A5E, GraphemeClass::ExtendIncb),
    (0x1A60, 0x1A60, GraphemeClass::Linker),
    (0x1A62, 0x1A62, GraphemeClass::ExtendIncb),
    (0x1A65, 0x1A6C, GraphemeClass::ExtendIncb),
    (0x1A6D, 0x1A72, GraphemeClass::SpacingMark),
    (0x1A73, 0x1A7C, GraphemeClass::ExtendIncb),
    (0x1A7F, 0x1A7F, GraphemeClass::ExtendIncb),
    (0x1AB0, 0x1ADD, GraphemeClass::ExtendIncb),
    (0x1AE0, 0x1AEB, GraphemeClass::ExtendIncb),
    (0x1B00, 0x1B03, GraphemeClass::ExtendIncb),
    (0x1B04, 0x1B04, GraphemeClass::SpacingMark),
    (0x1B0B, 0x1B0C, GraphemeClass::Consonant),
    (0x1B13, 0x1B33, GraphemeClass::Consonant),
    (0x1B34, 0x1B3D, GraphemeClass::ExtendIncb),
    (0x1B3E, 0x1B41, GraphemeClass::SpacingMark),
    (0x1B42, 0x1B43, GraphemeClass::ExtendIncb),
    (0x1B44, 0x1B44, GraphemeClass::Linker),
    (0x1B45, 0x1B4C, GraphemeClass::Consonant),
    (0x1B6B, 0x1B73, GraphemeClass::ExtendIncb),
    (0x1B80, 0x1B81, GraphemeClass::ExtendIncb),
    (0x1B82, 0x1B82, GraphemeClass::SpacingMark),
    (0x1B83, 0x1BA0, GraphemeClass::Consonant),
    (0x1BA1, 0x1BA1, GraphemeClass::SpacingMark),
    (0x1BA2, 0x1BA5, GraphemeClass::ExtendIncb),
    (0x1BA6, 0x1BA7, GraphemeClass::SpacingMark),
    (0x1BA8, 0x1BAA, GraphemeClass::ExtendIncb),
    (0x1BAB, 0x1BAB, GraphemeClass::Linker),
    (0x1BAC, 0x1BAD, GraphemeClass::ExtendIncb),
    (0x1BAE, 0x1BAF, GraphemeClass::Consonant),
    (0x1BBB, 0x1BBD, GraphemeClass::Consonant),
    (0x1BE6, 0x1BE6, GraphemeClass::ExtendIncb),
    (0x1BE7, 0x1BE7, GraphemeClass::SpacingMark),
    (0x1BE8, 0x1BE9, GraphemeClass::ExtendIncb),
    (0x1BEA, 0x1BEC, GraphemeClass::SpacingMark),
    (0x1BED, 0x1BED, GraphemeClass::ExtendIncb),
    (0x1BEE, 0x1BEE, GraphemeClass::SpacingMark),
    (0x1BEF, 0x1BF3, GraphemeClass::ExtendIncb),
    (0x1C24, 0x1C2B, GraphemeClass::SpacingMark),
    (0x1C2C, 0x1C33, GraphemeClass::ExtendIncb),
    (0x1C34, 0x1C35, GraphemeClass::SpacingMark),
    (0x1C36, 0x1C37, GraphemeClass::ExtendIncb),
    (0x1CD0, 0x1CD2, GraphemeClass::ExtendIncb),
    (0x1CD4, 0x1CE0, GraphemeClass::ExtendIncb),
    (0x1CE1, 0x1CE1, GraphemeClass::SpacingMark),
    (0x1CE2, 0x1CE8, GraphemeClass::ExtendIncb),
    (0x1CED, 0x1CED, GraphemeClass::ExtendIncb),
    (0x1CF4, 0x1CF4, GraphemeClass::ExtendIncb),
    (0x1CF7, 0x1CF7, GraphemeClass::SpacingMark),
    (0x1CF8, 0x1CF9, GraphemeClass::ExtendIncb),
    (0x1DC0, 0x1DFF, GraphemeClass::ExtendIncb),
    (0x200B, 0x200B, GraphemeClass::Control),
    (0x200C, 0x200C, GraphemeClass::Extend),
    (0x200D, 0x200D, GraphemeClass::Zwj),
    (0x200E, 0x200F, GraphemeClass::Control),
    (0x2028, 0x202E, GraphemeClass::Control),
    (0x203C, 0x203C, GraphemeClass::ExtPic),
    (0x2049, 0x2049, GraphemeClass::ExtPic),
    (0x2060, 0x206F, GraphemeClass::Control),
    (0x20D0, 0x20F0, GraphemeClass::ExtendIncb),
    (0x2122, 0x2122, GraphemeClass::ExtPic),
    (0x2139, 0x2139, GraphemeClass::ExtPic),
    (0x2194, 0x2199, GraphemeClass::ExtPic),
    (0x21A9, 0x21AA, GraphemeClass::ExtPic),
    (0x231A, 0x231B, GraphemeClass::ExtPic),
    (0x2328, 0x2328, GraphemeClass::ExtPic),
    (0x23CF, 0x23CF, GraphemeClass::ExtPic),
    (0x23E9, 0x23F3, GraphemeClass::ExtPic),
    (0x23F8, 0x23FA, GraphemeClass::ExtPic),
    (0x24C2, 0x24C2, GraphemeClass::ExtPic),
    (0x25AA, 0x25AB, GraphemeClass::ExtPic),
    (0x25B6, 0x25B6, GraphemeClass::ExtPic),
    (0x25C0, 0x25C0, GraphemeClass::ExtPic),
    (0x25FB, 0x25FE, GraphemeClass::ExtPic),
    (0x2600, 0x2604, GraphemeClass::ExtPic),
    (0x260E, 0x260E, GraphemeClass::ExtPic),
    (0x2611, 0x2611, GraphemeClass::ExtPic),
    (0x2614, 0x2615, GraphemeClass::ExtPic),
    (0x2618, 0x2618, GraphemeClass::ExtPic),
    (0x261D, 0x261D, GraphemeClass::ExtPic),
    (0x2620, 0x2620, GraphemeClass::ExtPic),
    (0x2622, 0x2623, GraphemeClass::ExtPic),
    (0x2626, 0x2626, GraphemeClass::ExtPic),
    (0x262A, 0x262A, GraphemeClass::ExtPic),
    (0x262E, 0x262F, GraphemeClass::ExtPic),
    (0x2638, 0x263A, GraphemeClass::ExtPic),
    (0x2640, 0x2640, GraphemeClass::ExtPic),
    (0x2642, 0x2642, GraphemeClass::ExtPic),
    (0x2648, 0x2653, GraphemeClass::ExtPic),
    (0x265F, 0x2660, GraphemeClass::ExtPic),
    (0x2663, 0x2663, GraphemeClass::ExtPic),
    (0x2665, 0x2666, GraphemeClass::ExtPic),
    (0x2668, 0x2668, GraphemeClass::ExtPic),
    (0x267B, 0x267B, GraphemeClass::ExtPic),
    (0x267E, 0x267F, GraphemeClass::ExtPic),
    (0x2692, 0x2697, GraphemeClass::ExtPic),
    (0x2699, 0x2699, GraphemeClass::ExtPic),
    (0x269B, 0x269C, GraphemeClass::ExtPic),
    (0x26A0, 0x26A1, GraphemeClass::ExtPic),
    (0x26A7, 0x26A7, GraphemeClass::ExtPic),
    (0x26AA, 0x26AB, GraphemeClass::ExtPic),
    (0x26B0, 0x26B1, GraphemeClass::ExtPic),
    (0x26BD, 0x26BE, GraphemeClass::ExtPic),
    (0x26C4, 0x26C5, GraphemeClass::ExtPic),
    (0x26C8, 0x26C8, GraphemeClass::ExtPic),
    (0x26CE, 0x26CF, GraphemeClass::ExtPic),
    (0x26D1, 0x26D1, GraphemeClass::ExtPic),
    (0x26D3, 0x26D4, GraphemeClass::ExtPic),
    (0x26E9, 0x26EA, GraphemeClass::ExtPic),
    (0x26F0, 0x26F5, GraphemeClass::ExtPic),
    (0x26F7, 0x26FA, GraphemeClass::ExtPic),
    (0x26FD, 0x26FD, GraphemeClass::ExtPic),
    (0x2702, 0x2702, GraphemeClass::ExtPic),
    (0x2705, 0x2705, GraphemeClass::ExtPic),
    (0x2708, 0x270D, GraphemeClass::ExtPic),
    (0x270F, 0x270F, GraphemeClass::ExtPic),
    (0x2712, 0x2712, GraphemeClass::ExtPic),
    (0x2714, 0x2714, GraphemeClass::ExtPic),
    (0x2716, 0x2716, GraphemeClass::ExtPic),
    (0x271D, 0x271D, GraphemeClass::ExtPic),
    (0x2721, 0x2721, GraphemeClass::ExtPic),
    (0x2728, 0x2728, GraphemeClass::ExtPic),
    (0x2733, 0x2734, GraphemeClass::ExtPic),
    (0x2744, 0x2744, GraphemeClass::ExtPic),
    (0x2747, 0x2747, GraphemeClass::ExtPic),
    (0x274C, 0x274C, GraphemeClass::ExtPic),
    (0x274E, 0x274E, GraphemeClass::ExtPic),
    (0x2753, 0x2755, GraphemeClass::ExtPic),
    (0x2757, 0x2757, GraphemeClass::ExtPic),
    (0x2763, 0x2764, GraphemeClass::ExtPic),
    (0x2795, 0x2797, GraphemeClass::ExtPic),
    (0x27A1, 0x27A1, GraphemeClass::ExtPic),
    (0x27B0, 0x27B0, GraphemeClass::ExtPic),
    (0x27BF, 0x27BF, GraphemeClass::ExtPic),
    (0x2934, 0x2935, GraphemeClass::ExtPic),
    (0x2B05, 0x2B07, GraphemeClass::ExtPic),
    (0x2B1B, 0x2B1C, GraphemeClass::ExtPic),
    (0x2B50, 0x2B50, GraphemeClass::ExtPic),
    (0x2B55, 0x2B55, GraphemeClass::ExtPic),
    (0x2CEF, 0x2CF1, GraphemeClass::ExtendIncb),
    (0x2D7F, 0x2D7F, GraphemeClass::ExtendIncb),
    (0x2DE0, 0x2DFF, GraphemeClass::ExtendIncb),
    (0x302A, 0x302F, GraphemeClass::ExtendIncb),
    (0x3030, 0x3030, GraphemeClass::ExtPic),
    (0x303D, 0x303D, GraphemeClass::ExtPic),
    (0x3099, 0x309A, GraphemeClass::ExtendIncb),
    (0x3297, 0x3297, GraphemeClass::ExtPic),
    (0x3299, 0x3299, GraphemeClass::ExtPic),
    (0xA66F, 0xA672, GraphemeClass::ExtendIncb),
    (0xA674, 0xA67D, GraphemeClass::ExtendIncb),
    (0xA69E, 0xA69F, GraphemeClass::ExtendIncb),
    (0xA6F0, 0xA6F1, GraphemeClass::ExtendIncb),
    (0xA802, 0xA802, GraphemeClass::ExtendIncb),
    (0xA806, 0xA806, GraphemeClass::ExtendIncb),
    (0xA80B, 0xA80B, GraphemeClass::ExtendIncb),
    (0xA823, 0xA824, GraphemeClass::SpacingMark),
    (0xA825, 0xA826, GraphemeClass::ExtendIncb),
    (0xA827, 0xA827, GraphemeClass::SpacingMark),
    (0xA82C, 0xA82C, GraphemeClass::ExtendIncb),
    (0xA880, 0xA881, GraphemeClass::SpacingMark),
    (0xA8B4, 0xA8C3, GraphemeClass::SpacingMark),
    (0xA8C4, 0xA8C5, GraphemeClass::ExtendIncb),
    (0xA8E0, 0xA8F1, GraphemeClass::ExtendIncb),
    (0xA8FF, 0xA8FF, GraphemeClass::ExtendIncb),
    (0xA926, 0xA92D, GraphemeClass::ExtendIncb),
    (0xA947, 0xA951, GraphemeClass::ExtendIncb),
    (0xA952, 0xA952, GraphemeClass::SpacingMark),
    (0xA953, 0xA953, GraphemeClass::ExtendIncb),
    (0xA960, 0xA97C, GraphemeClass::L),
    (0xA980, 0xA982, GraphemeClass::ExtendIncb),
    (0xA983, 0xA983, GraphemeClass::SpacingMark),
    (0xA989, 0xA98B, GraphemeClass::Consonant),
    (0xA98F, 0xA9B2, GraphemeClass::Consonant),
    (0xA9B3, 0xA9B3, GraphemeClass::ExtendIncb),
    (0xA9B4, 0xA9B5, GraphemeClass::SpacingMark),
    (0xA9B6, 0xA9B9, GraphemeClass::ExtendIncb),
    (0xA9BA, 0xA9BB, GraphemeClass::SpacingMark),
    (0xA9BC, 0xA9BD, GraphemeClass::ExtendIncb),
    (0xA9BE, 0xA9BF, GraphemeClass::SpacingMark),
    (0xA9C0, 0xA9C0, GraphemeClass::Linker),
    (0xA9E0, 0xA9E4, GraphemeClass::Consonant),
    (0xA9E5, 0xA9E5, GraphemeClass::ExtendIncb),
    (0xA9E7, 0xA9EF, GraphemeClass::Consonant),
    (0xA9FA, 0xA9FE, GraphemeClass::Consonant),
    (0xAA29, 0xAA2E, GraphemeClass::ExtendIncb),
    (0xAA2F, 0xAA30, GraphemeClass::SpacingMark),
    (0xAA31, 0xAA32, GraphemeClass::ExtendIncb),
    (0xAA33, 0xAA34, GraphemeClass::SpacingMark),
    (0xAA35, 0xAA36, GraphemeClass::ExtendIncb),
    (0xAA43, 0xAA43, GraphemeClass::ExtendIncb),
    (0xAA4C, 0xAA4C, GraphemeClass::ExtendIncb),
    (0xAA4D, 0xAA4D, GraphemeClass::SpacingMark),
    (0xAA60, 0xAA6F, GraphemeClass::Consonant),
    (0xAA71, 0xAA73, GraphemeClass::Consonant),
    (0xAA7A, 0xAA7A, GraphemeClass::Consonant),
    (0xAA7C, 0xAA7C, GraphemeClass::ExtendIncb),
    (0xAA7E, 0xAA7F, GraphemeClass::Consonant),
    (0xAAB0, 0xAAB0, GraphemeClass::ExtendIncb),
    (0xAAB2, 0xAAB4, GraphemeClass::ExtendIncb),
    (0xAAB7, 0xAAB8, GraphemeClass::ExtendIncb),
    (0xAABE, 0xAABF, GraphemeClass::ExtendIncb),
    (0xAAC1, 0xAAC1, GraphemeClass::ExtendIncb),
    (0xAAE0, 0xAAEA, GraphemeClass::Consonant),
    (0xAAEB, 0xAAEB, GraphemeClass::SpacingMark),
    (0xAAEC, 0xAAED, GraphemeClass::ExtendIncb),
    (0xAAEE, 0xAAEF, GraphemeClass::SpacingMark),
    (0xAAF5, 0xAAF5, GraphemeClass::SpacingMark),
    (0xAAF6, 0xAAF6, GraphemeClass::Linker),
    (0xABC0, 0xABDA, GraphemeClass::Consonant),
    (0xABE3, 0xABE4, GraphemeClass::SpacingMark),
    (0xABE5, 0xABE5, GraphemeClass::ExtendIncb),
    (0xABE6, 0xABE7, GraphemeClass::SpacingMark),
    (0xABE8, 0xABE8, GraphemeClass::ExtendIncb),
    (0xABE9, 0xABEA, GraphemeClass::SpacingMark),
    (0xABEC, 0xABEC, GraphemeClass::SpacingMark),
    (0xABED, 0xABED, GraphemeClass::ExtendIncb),
    (0xAC00, 0xAC00, GraphemeClass::LV),
    (0xAC01, 0xAC1B, GraphemeClass::LVT),
    (0xAC1C, 0xAC1C, GraphemeClass::LV),
    (0xAC1D, 0xAC37, GraphemeClass::LVT),
    (0xAC38, 0xAC38, GraphemeClass::LV),
    (0xAC39, 0xAC53, GraphemeClass::LVT),
    (0xAC54, 0xAC54, GraphemeClass::LV),
    (0xAC55, 0xAC6F, GraphemeClass::LVT),
    (0xAC70, 0xAC70, GraphemeClass::LV),
    (0xAC71, 0xAC8B, GraphemeClass::LVT),
    (0xAC8C, 0xAC8C, GraphemeClass::LV),
    (0xAC8D, 0xACA7, GraphemeClass::LVT),
    (0xACA8, 0xACA8, GraphemeClass::LV),
    (0xACA9, 0xACC3, GraphemeClass::LVT),
    (0xACC4, 0xACC4, GraphemeClass::LV),
    (0xACC5, 0xACDF, GraphemeClass::LVT),
    (0xACE0, 0xACE0, GraphemeClass::LV),
    (0xACE1, 0xACFB, GraphemeClass::LVT),
    (0xACFC, 0xACFC, GraphemeClass::LV),
    (0xACFD, 0xAD17, GraphemeClass::LVT),
    (0xAD18, 0xAD18, GraphemeClass::LV),
    (0xAD19, 0xAD33, GraphemeClass::LVT),
    (0xAD34, 0xAD34, GraphemeClass::LV),
    (0xAD35, 0xAD4F, GraphemeClass::LVT),
    (0xAD50, 0xAD50, GraphemeClass::LV),
    (0xAD51, 0xAD6B, GraphemeClass::LVT),
    (0xAD6C, 0xAD6C, GraphemeClass::LV),
    (0xAD6D, 0xAD87, GraphemeClass::LVT),
    (0xAD88, 0xAD88, GraphemeClass::LV),
    (0xAD89, 0xADA3, GraphemeClass::LVT),
    (0xADA4, 0xADA4, GraphemeClass::LV),
    (0xADA5, 0xADBF, GraphemeClass::LVT),
    (0xADC0, 0xADC0, GraphemeClass::LV),
    (0xADC1, 0xADDB, GraphemeClass::LVT),
    (0xADDC, 0xADDC, GraphemeClass::LV),
    (0xADDD, 0xADF7, GraphemeClass::LVT),
    (0xADF8, 0xADF8, GraphemeClass::LV),
    (0xADF9, 0xAE13, GraphemeClass::LVT),
    (0xAE14, 0xAE14, GraphemeClass::LV),
    (0xAE15, 0xAE2F, GraphemeClass::LVT),
    (0xAE30, 0xAE30, GraphemeClass::LV),
    (0xAE31, 0xAE4B, GraphemeClass::LVT),
    (0xAE4C, 0xAE4C, GraphemeClass::LV),
    (0xAE4D, 0xAE67, GraphemeClass::LVT),
    (0xAE68, 0xAE68, GraphemeClass::LV),
    (0xAE69, 0xAE83, GraphemeClass::LVT),
    (0xAE84, 0xAE84, GraphemeClass::LV),
    (0xAE85, 0xAE9F, GraphemeClass::LVT),
    (0xAEA0, 0xAEA0, GraphemeClass::LV),
    (0xAEA1, 0xAEBB, GraphemeClass::LVT),
    (0xAEBC, 0xAEBC, GraphemeClass::LV),
    (0xAEBD, 0xAED7, GraphemeClass::LVT),
    (0xAED8, 0xAED8, GraphemeClass::LV),
    (0xAED9, 0xAEF3, GraphemeClass::LVT),
    (0xAEF4, 0xAEF4, GraphemeClass::LV),
    (0xAEF5, 0xAF0F, GraphemeClass::LVT),
    (0xAF10, 0xAF10, GraphemeClass::LV),
    (0xAF11, 0xAF2B, GraphemeClass::LVT),
    (0xAF2C, 0xAF2C, GraphemeClass::LV),
    (0xAF2D, 0xAF47, GraphemeClass::LVT),
    (0xAF48, 0xAF48, GraphemeClass::LV),
    (0xAF49, 0xAF63, GraphemeClass::LVT),
    (0xAF64, 0xAF64, GraphemeClass::LV),
    (0xAF65, 0xAF7F, GraphemeClass::LVT),
    (0xAF80, 0xAF80, GraphemeClass::LV),
    (0xAF81, 0xAF9B, GraphemeClass::LVT),
    (0xAF9C, 0xAF9C, GraphemeClass::LV),
    (0xAF9D, 0xAFB7, GraphemeClass::LVT),
    (0xAFB8, 0xAFB8, GraphemeClass::LV),
    (0xAFB9, 0xAFD3, GraphemeClass::LVT),
    (0xAFD4, 0xAFD4, GraphemeClass::LV),
    (0xAFD5, 0xAFEF, GraphemeClass::LVT),
    (0xAFF0, 0xAFF0, GraphemeClass::LV),
    (0xAFF1, 0xB00B, GraphemeClass::LVT),
    (0xB00C, 0xB00C, GraphemeClass::LV),
    (0xB00D, 0xB027, GraphemeClass::LVT),
    (0xB028, 0xB028, GraphemeClass::LV),
    (0xB029, 0xB043, GraphemeClass::LVT),
    (0xB044, 0xB044, GraphemeClass::LV),
    (0xB045, 0xB05F, GraphemeClass::LVT),
    (0xB060, 0xB060, GraphemeClass::LV),
    (0xB061, 0xB07B, GraphemeClass::LVT),
    (0xB07C, 0xB07C, GraphemeClass::LV),
    (0xB07D, 0xB097, GraphemeClass::LVT),
    (0xB098, 0xB098, GraphemeClass::LV),
    (0xB099, 0xB0B3, GraphemeClass::LVT),
    (0xB0B4, 0xB0B4, GraphemeClass::LV),
    (0xB0B5, 0xB0CF, GraphemeClass::LVT),
    (0xB0D0, 0xB0D0, GraphemeClass::LV),
    (0xB0D1, 0xB0EB, GraphemeClass::LVT),
    (0xB0EC, 0xB0EC, GraphemeClass::LV),
    (0xB0ED, 0xB107, GraphemeClass::LVT),
    (0xB108, 0xB108, GraphemeClass::LV),
    (0xB109, 0xB123, GraphemeClass::LVT),
    (0xB124, 0xB124, GraphemeClass::LV),
    (0xB125, 0xB13F, GraphemeClass::LVT),
    (0xB140, 0xB140, GraphemeClass::LV),
    (0xB141, 0xB15B, GraphemeClass::LVT),
    (0xB15C, 0xB15C, GraphemeClass::LV),
    (0xB15D, 0xB177, GraphemeClass::LVT),
    (0xB178, 0xB178, GraphemeClass::LV),
    (0xB179, 0xB193, GraphemeClass::LVT),
    (0xB194, 0xB194, GraphemeClass::LV),
    (0xB195, 0xB1AF, GraphemeClass::LVT),
    (0xB1B0, 0xB1B0, GraphemeClass::LV),
    (0xB1B1, 0xB1CB, GraphemeClass::LVT),
    (0xB1CC, 0xB1CC, GraphemeClass::LV),
    (0xB1CD, 0xB1E7, GraphemeClass::LVT),
    (0xB1E8, 0xB1E8, GraphemeClass::LV),
    (0xB1E9, 0xB203, GraphemeClass::LVT),
    (0xB204, 0xB204, GraphemeClass::LV),
    (0xB205, 0xB21F, GraphemeClass::LVT),
    (0xB220, 0xB220, GraphemeClass::LV),
    (0xB221, 0xB23B, GraphemeClass::LVT),
    (0xB23C, 0xB23C, GraphemeClass::LV),
    (0xB23D, 0xB257, GraphemeClass::LVT),
    (0xB258, 0xB258, GraphemeClass::LV),
    (0xB259, 0xB273, GraphemeClass::LVT),
    (0xB274, 0xB274, GraphemeClass::LV),
    (0xB275, 0xB28F, GraphemeClass::LVT),
    (0xB290, 0xB290, GraphemeClass::LV),
    (0xB291, 0xB2AB, GraphemeClass::LVT),
    (0xB2AC, 0xB2AC, GraphemeClass::LV),
    (0xB2AD, 0xB2C7, GraphemeClass::LVT),
    (0xB2C8, 0xB2C8, GraphemeClass::LV),
    (0xB2C9, 0xB2E3, GraphemeClass::LVT),
    (0xB2E4, 0xB2E4, GraphemeClass::LV),
    (0xB2E5, 0xB2FF, GraphemeClass::LVT),
    (0xB300, 0xB300, GraphemeClass::LV),
    (0xB301, 0xB31B, GraphemeClass::LVT),
    (0xB31C, 0xB31C, GraphemeClass::LV),
    (0xB31D, 0xB337, GraphemeClass::LVT),
    (0xB338, 0xB338, GraphemeClass::LV),
    (0xB339, 0xB353, GraphemeClass::LVT),
    (0xB354, 0xB354, GraphemeClass::LV),
    (0xB355, 0xB36F, GraphemeClass::LVT),
    (0xB370, 0xB370, GraphemeClass::LV),
    (0xB371, 0xB38B, GraphemeClass::LVT),
    (0xB38C, 0xB38C, GraphemeClass::LV),
    (0xB38D, 0xB3A7, GraphemeClass::LVT),
    (0xB3A8, 0xB3A8, GraphemeClass::LV),
    (0xB3A9, 0xB3C3, GraphemeClass::LVT),
    (0xB3C4, 0xB3C4, GraphemeClass::LV),
    (0xB3C5, 0xB3DF, GraphemeClass::LVT),
    (0xB3E0, 0xB3E0, GraphemeClass::LV),
    (0xB3E1, 0xB3FB, GraphemeClass::LVT),
    (0xB3FC, 0xB3FC, GraphemeClass::LV),
    (0xB3FD, 0xB417, GraphemeClass::LVT),
    (0xB418, 0xB418, GraphemeClass::LV),
    (0xB419, 0xB433, GraphemeClass::LVT),
    (0xB434, 0xB434, GraphemeClass::LV),
    (0xB435, 0xB44F, GraphemeClass::LVT),
    (0xB450, 0xB450, GraphemeClass::LV),
    (0xB451, 0xB46B, GraphemeClass::LVT),
    (0xB46C, 0xB46C, GraphemeClass::LV),
    (0xB46D, 0xB487, GraphemeClass::LVT),
    (0xB488, 0xB488, GraphemeClass::LV),
    (0xB489, 0xB4A3, GraphemeClass::LVT),
    (0xB4A4, 0xB4A4, GraphemeClass::LV),
    (0xB4A5, 0xB4BF, GraphemeClass::LVT),
    (0xB4C0, 0xB4C0, GraphemeClass::LV),
    (0xB4C1, 0xB4DB, GraphemeClass::LVT),
    (0xB4DC, 0xB4DC, GraphemeClass::LV),
    (0xB4DD, 0xB4F7, GraphemeClass::LVT),
    (0xB4F8, 0xB4F8, GraphemeClass::LV),
    (0xB4F9, 0xB513, GraphemeClass::LVT),
    (0xB514, 0xB514, GraphemeClass::LV),
    (0xB515, 0xB52F, GraphemeClass::LVT),
    (0xB530, 0xB530, GraphemeClass::LV),
    (0xB531, 0xB54B, GraphemeClass::LVT),
    (0xB54C, 0xB54C, GraphemeClass::LV),
    (0xB54D, 0xB567, GraphemeClass::LVT),
    (0xB568, 0xB568, GraphemeClass::LV),
    (0xB569, 0xB583, GraphemeClass::LVT),
    (0xB584, 0xB584, GraphemeClass::LV),
    (0xB585, 0xB59F, GraphemeClass::LVT),
    (0xB5A0, 0xB5A0, GraphemeClass::LV),
    (0xB5A1, 0xB5BB, GraphemeClass::LVT),
    (0xB5BC, 0xB5BC, GraphemeClass::LV),
    (0xB5BD, 0xB5D7, GraphemeClass::LVT),
    (0xB5D8, 0xB5D8, GraphemeClass::LV),
    (0xB5D9, 0xB5F3, GraphemeClass::LVT),
    (0xB5F4, 0xB5F4, GraphemeClass::LV),
    (0xB5F5, 0xB60F, GraphemeClass::LVT),
    (0xB610, 0xB610, GraphemeClass::LV),
    (0xB611, 0xB62B, GraphemeClass::LVT),
    (0xB62C, 0xB62C, GraphemeClass::LV),
    (0xB62D, 0xB647, GraphemeClass::LVT),
    (0xB648, 0xB648, GraphemeClass::LV),
    (0xB649, 0xB663, GraphemeClass::LVT),
    (0xB664, 0xB664, GraphemeClass::LV),
    (0xB665, 0xB67F, GraphemeClass::LVT),
    (0xB680, 0xB680, GraphemeClass::LV),
    (0xB681, 0xB69B, GraphemeClass::LVT),
    (0xB69C, 0xB69C, GraphemeClass::LV),
    (0xB69D, 0xB6B7, GraphemeClass::LVT),
    (0xB6B8, 0xB6B8, GraphemeClass::LV),
    (0xB6B9, 0xB6D3, GraphemeClass::LVT),
    (0xB6D4, 0xB6D4, GraphemeClass::LV),
    (0xB6D5, 0xB6EF, GraphemeClass::LVT),
    (0xB6F0, 0xB6F0, GraphemeClass::LV),
    (0xB6F1, 0xB70B, GraphemeClass::LVT),
    (0xB70C, 0xB70C, GraphemeClass::LV),
    (0xB70D, 0xB727, GraphemeClass::LVT),
    (0xB728, 0xB728, GraphemeClass::LV),
    (0xB729, 0xB743, GraphemeClass::LVT),
    (0xB744, 0xB744, GraphemeClass::LV),
    (0xB745, 0xB75F, GraphemeClass::LVT),
    (0xB760, 0xB760, GraphemeClass::LV),
    (0xB761, 0xB77B, GraphemeClass::LVT),
    (0xB77C, 0xB77C, GraphemeClass::LV),
    (0xB77D, 0xB797, GraphemeClass::LVT),
    (0xB798, 0xB798, GraphemeClass::LV),
    (0xB799, 0xB7B3, GraphemeClass::LVT),
    (0xB7B4, 0xB7B4, GraphemeClass::LV),
    (0xB7B5, 0xB7CF, GraphemeClass::LVT),
    (0xB7D0, 0xB7D0, GraphemeClass::LV),
    (0xB7D1, 0xB7EB, GraphemeClass::LVT),
    (0xB7EC, 0xB7EC, GraphemeClass::LV),
    (0xB7ED, 0xB807, GraphemeClass::LVT),
    (0xB808, 0xB808, GraphemeClass::LV),
    (0xB809, 0xB823, GraphemeClass::LVT),
    (0xB824, 0xB824, GraphemeClass::LV),
    (0xB825, 0xB83F, GraphemeClass::LVT),
    (0xB840, 0xB840, GraphemeClass::LV),
    (0xB841, 0xB85B, GraphemeClass::LVT),
    (0xB85C, 0xB85C, GraphemeClass::LV),
    (0xB85D, 0xB877, GraphemeClass::LVT),
    (0xB878, 0xB878, GraphemeClass::LV),
    (0xB879, 0xB893, GraphemeClass::LVT),
    (0xB894, 0xB894, GraphemeClass::LV),
    (0xB895, 0xB8AF, GraphemeClass::LVT),
    (0xB8B0, 0xB8B0, GraphemeClass::LV),
    (0xB8B1, 0xB8CB, GraphemeClass::LVT),
    (0xB8CC, 0xB8CC, GraphemeClass::LV),
    (0xB8CD, 0xB8E7, GraphemeClass::LVT),
    (0xB8E8, 0xB8E8, GraphemeClass::LV),
    (0xB8E9, 0xB903, GraphemeClass::LVT),
    (0xB904, 0xB904, GraphemeClass::LV),
    (0xB905, 0xB91F, GraphemeClass::LVT),
    (0xB920, 0xB920, GraphemeClass::LV),
    (0xB921, 0xB93B, GraphemeClass::LVT),
    (0xB93C, 0xB93C, GraphemeClass::LV),
    (0xB93D, 0xB957, GraphemeClass::LVT),
    (0xB958, 0xB958, GraphemeClass::LV),
    (0xB959, 0xB973, GraphemeClass::LVT),
    (0xB974, 0xB974, GraphemeClass::LV),
    (0xB975, 0xB98F, GraphemeClass::LVT),
    (0xB990, 0xB990, GraphemeClass::LV),
    (0xB991, 0xB9AB, GraphemeClass::LVT),
    (0xB9AC, 0xB9AC, GraphemeClass::LV),
    (0xB9AD, 0xB9C7, GraphemeClass::LVT),
    (0xB9C8, 0xB9C8, GraphemeClass::LV),
    (0xB9C9, 0xB9E3, GraphemeClass::LVT),
    (0xB9E4, 0xB9E4, GraphemeClass::LV),
    (0xB9E5, 0xB9FF, GraphemeClass::LVT),
    (0xBA00, 0xBA00, GraphemeClass::LV),
    (0xBA01, 0xBA1B, GraphemeClass::LVT),
    (0xBA1C, 0xBA1C, GraphemeClass::LV),
    (0xBA1D, 0xBA37, GraphemeClass::LVT),
    (0xBA38, 0xBA38, GraphemeClass::LV),
    (0xBA39, 0xBA53, GraphemeClass::LVT),
    (0xBA54, 0xBA54, GraphemeClass::LV),
    (0xBA55, 0xBA6F, GraphemeClass::LVT),
    (0xBA70, 0xBA70, GraphemeClass::LV),
    (0xBA71, 0xBA8B, GraphemeClass::LVT),
    (0xBA8C, 0xBA8C, GraphemeClass::LV),
    (0xBA8D, 0xBAA7, GraphemeClass::LVT),
    (0xBAA8, 0xBAA8, GraphemeClass::LV),
    (0xBAA9, 0xBAC3, GraphemeClass::LVT),
    (0xBAC4, 0xBAC4, GraphemeClass::LV),
    (0xBAC5, 0xBADF, GraphemeClass::LVT),
    (0xBAE0, 0xBAE0, GraphemeClass::LV),
    (0xBAE1, 0xBAFB, GraphemeClass::LVT),
    (0xBAFC, 0xBAFC, GraphemeClass::LV),
    (0xBAFD, 0xBB17, GraphemeClass::LVT),
    (0xBB18, 0xBB18, GraphemeClass::LV),
    (0xBB19, 0xBB33, GraphemeClass::LVT),
    (0xBB34, 0xBB34, GraphemeClass::LV),
    (0xBB35, 0xBB4F, GraphemeClass::LVT),
    (0xBB50, 0xBB50, GraphemeClass::LV),
    (0xBB51, 0xBB6B, GraphemeClass::LVT),
    (0xBB6C, 0xBB6C, GraphemeClass::LV),
    (0xBB6D, 0xBB87, GraphemeClass::LVT),
    (0xBB88, 0xBB88, GraphemeClass::LV),
    (0xBB89, 0xBBA3, GraphemeClass::LVT),
    (0xBBA4, 0xBBA4, GraphemeClass::LV),
    (0xBBA5, 0xBBBF, GraphemeClass::LVT),
    (0xBBC0, 0xBBC0, GraphemeClass::LV),
    (0xBBC1, 0xBBDB, GraphemeClass::LVT),
    (0xBBDC, 0xBBDC, GraphemeClass::LV),
    (0xBBDD, 0xBBF7, GraphemeClass::LVT),
    (0xBBF8, 0xBBF8, GraphemeClass::LV),
    (0xBBF9, 0xBC13, GraphemeClass::LVT),
    (0xBC14, 0xBC14, GraphemeClass::LV),
    (0xBC15, 0xBC2F, GraphemeClass::LVT),
    (0xBC30, 0xBC30, GraphemeClass::LV),
    (0xBC31, 0xBC4B, GraphemeClass::LVT),
    (0xBC4C, 0xBC4C, GraphemeClass::LV),
    (0xBC4D, 0xBC67, GraphemeClass::LVT),
    (0xBC68, 0xBC68, GraphemeClass::LV),
    (0xBC69, 0xBC83, GraphemeClass::LVT),
    (0xBC84, 0xBC84, GraphemeClass::LV),
    (0xBC85, 0xBC9F, GraphemeClass::LVT),
    (0xBCA0, 0xBCA0, GraphemeClass::LV),
    (0xBCA1, 0xBCBB, GraphemeClass::LVT),
    (0xBCBC, 0xBCBC, GraphemeClass::LV),
    (0xBCBD, 0xBCD7, GraphemeClass::LVT),
    (0xBCD8, 0xBCD8, GraphemeClass::LV),
    (0xBCD9, 0xBCF3, GraphemeClass::LVT),
    (0xBCF4, 0xBCF4, GraphemeClass::LV),
    (0xBCF5, 0xBD0F, GraphemeClass::LVT),
    (0xBD10, 0xBD10, GraphemeClass::LV),
    (0xBD11, 0xBD2B, GraphemeClass::LVT),
    (0xBD2C, 0xBD2C, GraphemeClass::LV),
    (0xBD2D, 0xBD47, GraphemeClass::LVT),
    (0xBD48, 0xBD48, GraphemeClass::LV),
    (0xBD49, 0xBD63, GraphemeClass::LVT),
    (0xBD64, 0xBD64, GraphemeClass::LV),
    (0xBD65, 0xBD7F, GraphemeClass::LVT),
    (0xBD80, 0xBD80, GraphemeClass::LV),
    (0xBD81, 0xBD9B, GraphemeClass::LVT),
    (0xBD9C, 0xBD9C, GraphemeClass::LV),
    (0xBD9D, 0xBDB7, GraphemeClass::LVT),
    (0xBDB8, 0xBDB8, GraphemeClass::LV),
    (0xBDB9, 0xBDD3, GraphemeClass::LVT),
    (0xBDD4, 0xBDD4, GraphemeClass::LV),
    (0xBDD5, 0xBDEF, GraphemeClass::LVT),
    (0xBDF0, 0xBDF0, GraphemeClass::LV),
    (0xBDF1, 0xBE0B, GraphemeClass::LVT),
    (0xBE0C, 0xBE0C, GraphemeClass::LV),
    (0xBE0D, 0xBE27, GraphemeClass::LVT),
    (0xBE28, 0xBE28, GraphemeClass::LV),
    (0xBE29, 0xBE43, GraphemeClass::LVT),
    (0xBE44, 0xBE44, GraphemeClass::LV),
    (0xBE45, 0xBE5F, GraphemeClass::LVT),
    (0xBE60, 0xBE60, GraphemeClass::LV),
    (0xBE61, 0xBE7B, GraphemeClass::LVT),
    (0xBE7C, 0xBE7C, GraphemeClass::LV),
    (0xBE7D, 0xBE97, GraphemeClass::LVT),
    (0xBE98, 0xBE98, GraphemeClass::LV),
    (0xBE99, 0xBEB3, GraphemeClass::LVT),
    (0xBEB4, 0xBEB4, GraphemeClass::LV),
    (0xBEB5, 0xBECF, GraphemeClass::LVT),
    (0xBED0, 0xBED0, GraphemeClass::LV),
    (0xBED1, 0xBEEB, GraphemeClass::LVT),
    (0xBEEC, 0xBEEC, GraphemeClass::LV),
    (0xBEED, 0xBF07, GraphemeClass::LVT),
    (0xBF08, 0xBF08, GraphemeClass::LV),
    (0xBF09, 0xBF23, GraphemeClass::LVT),
    (0xBF24, 0xBF24, GraphemeClass::LV),
    (0xBF25, 0xBF3F, GraphemeClass::LVT),
    (0xBF40, 0xBF40, GraphemeClass::LV),
    (0xBF41, 0xBF5B, GraphemeClass::LVT),
    (0xBF5C, 0xBF5C, GraphemeClass::LV),
    (0xBF5D, 0xBF77, GraphemeClass::LVT),
    (0xBF78, 0xBF78, GraphemeClass::LV),
    (0xBF79, 0xBF93, GraphemeClass::LVT),
    (0xBF94, 0xBF94, GraphemeClass::LV),
    (0xBF95, 0xBFAF, GraphemeClass::LVT),
    (0xBFB0, 0xBFB0, GraphemeClass::LV),
    (0xBFB1, 0xBFCB, GraphemeClass::LVT),
    (0xBFCC, 0xBFCC, GraphemeClass::LV),
    (0xBFCD, 0xBFE7, GraphemeClass::LVT),
    (0xBFE8, 0xBFE8, GraphemeClass::LV),
    (0xBFE9, 0xC003, GraphemeClass::LVT),
    (0xC004, 0xC004, GraphemeClass::LV),
    (0xC005, 0xC01F, GraphemeClass::LVT),
    (0xC020, 0xC020, GraphemeClass::LV),
    (0xC021, 0xC03B, GraphemeClass::LVT),
    (0xC03C, 0xC03C, GraphemeClass::LV),
    (0xC03D, 0xC057, GraphemeClass::LVT),
    (0xC058, 0xC058, GraphemeClass::LV),
    (0xC059, 0xC073, GraphemeClass::LVT),
    (0xC074, 0xC074, GraphemeClass::LV),
    (0xC075, 0xC08F, GraphemeClass::LVT),
    (0xC090, 0xC090, GraphemeClass::LV),
    (0xC091, 0xC0AB, GraphemeClass::LVT),
    (0xC0AC, 0xC0AC, GraphemeClass::LV),
    (0xC0AD, 0xC0C7, GraphemeClass::LVT),
    (0xC0C8, 0xC0C8, GraphemeClass::LV),
    (0xC0C9, 0xC0E3, GraphemeClass::LVT),
    (0xC0E4, 0xC0E4, GraphemeClass::LV),
    (0xC0E5, 0xC0FF, GraphemeClass::LVT),
    (0xC100, 0xC100, GraphemeClass::LV),
    (0xC101, 0xC11B, GraphemeClass::LVT),
    (0xC11C, 0xC11C, GraphemeClass::LV),
    (0xC11D, 0xC137, GraphemeClass::LVT),
    (0xC138, 0xC138, GraphemeClass::LV),
    (0xC139, 0xC153, GraphemeClass::LVT),
    (0xC154, 0xC154, GraphemeClass::LV),
    (0xC155, 0xC16F, GraphemeClass::LVT),
    (0xC170, 0xC170, GraphemeClass::LV),
    (0xC171, 0xC18B, GraphemeClass::LVT),
    (0xC18C, 0xC18C, GraphemeClass::LV),
    (0xC18D, 0xC1A7, GraphemeClass::LVT),
    (0xC1A8, 0xC1A8, GraphemeClass::LV),
    (0xC1A9, 0xC1C3, GraphemeClass::LVT),
    (0xC1C4, 0xC1C4, GraphemeClass::LV),
    (0xC1C5, 0xC1DF, GraphemeClass::LVT),
    (0xC1E0, 0xC1E0, GraphemeClass::LV),
    (0xC1E1, 0xC1FB, GraphemeClass::LVT),
    (0xC1FC, 0xC1FC, GraphemeClass::LV),
    (0xC1FD, 0xC217, GraphemeClass::LVT),
    (0xC218, 0xC218, GraphemeClass::LV),
    (0xC219, 0xC233, GraphemeClass::LVT),
    (0xC234, 0xC234, GraphemeClass::LV),
    (0xC235, 0xC24F, GraphemeClass::LVT),
    (0xC250, 0xC250, GraphemeClass::LV),
    (0xC251, 0xC26B, GraphemeClass::LVT),
    (0xC26C, 0xC26C, GraphemeClass::LV),
    (0xC26D, 0xC287, GraphemeClass::LVT),
    (0xC288, 0xC288, GraphemeClass::LV),
    (0xC289, 0xC2A3, GraphemeClass::LVT),
    (0xC2A4, 0xC2A4, GraphemeClass::LV),
    (0xC2A5, 0xC2BF, GraphemeClass::LVT),
    (0xC2C0, 0xC2C0, GraphemeClass::LV),
    (0xC2C1, 0xC2DB, GraphemeClass::LVT),
    (0xC2DC, 0xC2DC, GraphemeClass::LV),
    (0xC2DD, 0xC2F7, GraphemeClass::LVT),
    (0xC2F8, 0xC2F8, GraphemeClass::LV),
    (0xC2F9, 0xC313, GraphemeClass::LVT),
    (0xC314, 0xC314, GraphemeClass::LV),
    (0xC315, 0xC32F, GraphemeClass::LVT),
    (0xC330, 0xC330, GraphemeClass::LV),
    (0xC331, 0xC34B, GraphemeClass::LVT),
    (0xC34C, 0xC34C, GraphemeClass::LV),
    (0xC34D, 0xC367, GraphemeClass::LVT),
    (0xC368, 0xC368, GraphemeClass::LV),
    (0xC369, 0xC383, GraphemeClass::LVT),
    (0xC384, 0xC384, GraphemeClass::LV),
    (0xC385, 0xC39F, GraphemeClass::LVT),
    (0xC3A0, 0xC3A0, GraphemeClass::LV),
    (0xC3A1, 0xC3BB, GraphemeClass::LVT),
    (0xC3BC, 0xC3BC, GraphemeClass::LV),
    (0xC3BD, 0xC3D7, GraphemeClass::LVT),
    (0xC3D8, 0xC3D8, GraphemeClass::LV),
    (0xC3D9, 0xC3F3, GraphemeClass::LVT),
    (0xC3F4, 0xC3F4, GraphemeClass::LV),
    (0xC3F5, 0xC40F, GraphemeClass::LVT),
    (0xC410, 0xC410, GraphemeClass::LV),
    (0xC411, 0xC42B, GraphemeClass::LVT),
    (0xC42C, 0xC42C, GraphemeClass::LV),
    (0xC42D, 0xC447, GraphemeClass::LVT),
    (0xC448, 0xC448, GraphemeClass::LV),
    (0xC449, 0xC463, GraphemeClass::LVT),
    (0xC464, 0xC464, GraphemeClass::LV),
    (0xC465, 0xC47F, GraphemeClass::LVT),
    (0xC480, 0xC480, GraphemeClass::LV),
    (0xC481, 0xC49B, GraphemeClass::LVT),
    (0xC49C, 0xC49C, GraphemeClass::LV),
    (0xC49D, 0xC4B7, GraphemeClass::LVT),
    (0xC4B8, 0xC4B8, GraphemeClass::LV),
    (0xC4B9, 0xC4D3, GraphemeClass::LVT),
    (0xC4D4, 0xC4D4, GraphemeClass::LV),
    (0xC4D5, 0xC4EF, GraphemeClass::LVT),
    (0xC4F0, 0xC4F0, GraphemeClass::LV),
    (0xC4F1, 0xC50B, GraphemeClass::LVT),
    (0xC50C, 0xC50C, GraphemeClass::LV),
    (0xC50D, 0xC527, GraphemeClass::LVT),
    (0xC528, 0xC528, GraphemeClass::LV),
    (0xC529, 0xC543, GraphemeClass::LVT),
    (0xC544, 0xC544, GraphemeClass::LV),
    (0xC545, 0xC55F, GraphemeClass::LVT),
    (0xC560, 0xC560, GraphemeClass::LV),
    (0xC561, 0xC57B, GraphemeClass::LVT),
    (0xC57C, 0xC57C, GraphemeClass::LV),
    (0xC57D, 0xC597, GraphemeClass::LVT),
    (0xC598, 0xC598, GraphemeClass::LV),
    (0xC599, 0xC5B3, GraphemeClass::LVT),
    (0xC5B4, 0xC5B4, GraphemeClass::LV),
    (0xC5B5, 0xC5CF, GraphemeClass::LVT),
    (0xC5D0, 0xC5D0, GraphemeClass::LV),
    (0xC5D1, 0xC5EB, GraphemeClass::LVT),
    (0xC5EC, 0xC5EC, GraphemeClass::LV),
    (0xC5ED, 0xC607, GraphemeClass::LVT),
    (0xC608, 0xC608, GraphemeClass::LV),
    (0xC609, 0xC623, GraphemeClass::LVT),
    (0xC624, 0xC624, GraphemeClass::LV),
    (0xC625, 0xC63F, GraphemeClass::LVT),
    (0xC640, 0xC640, GraphemeClass::LV),
    (0xC641, 0xC65B, GraphemeClass::LVT),
    (0xC65C, 0xC65C, GraphemeClass::LV),
    (0xC65D, 0xC677, GraphemeClass::LVT),
    (0xC678, 0xC678, GraphemeClass::LV),
    (0xC679, 0xC693, GraphemeClass::LVT),
    (0xC694, 0xC694, GraphemeClass::LV),
    (0xC695, 0xC6AF, GraphemeClass::LVT),
    (0xC6B0, 0xC6B0, GraphemeClass::LV),
    (0xC6B1, 0xC6CB, GraphemeClass::LVT),
    (0xC6CC, 0xC6CC, GraphemeClass::LV),
    (0xC6CD, 0xC6E7, GraphemeClass::LVT),
    (0xC6E8, 0xC6E8, GraphemeClass::LV),
    (0xC6E9, 0xC703, GraphemeClass::LVT),
    (0xC704, 0xC704, GraphemeClass::LV),
    (0xC705, 0xC71F, GraphemeClass::LVT),
    (0xC720, 0xC720, GraphemeClass::LV),
    (0xC721, 0xC73B, GraphemeClass::LVT),
    (0xC73C, 0xC73C, GraphemeClass::LV),
    (0xC73D, 0xC757, GraphemeClass::LVT),
    (0xC758, 0xC758, GraphemeClass::LV),
    (0xC759, 0xC773, GraphemeClass::LVT),
    (0xC774, 0xC774, GraphemeClass::LV),
    (0xC775, 0xC78F, GraphemeClass::LVT),
    (0xC790, 0xC790, GraphemeClass::LV),
    (0xC791, 0xC7AB, GraphemeClass::LVT),
    (0xC7AC, 0xC7AC, GraphemeClass::LV),
    (0xC7AD, 0xC7C7, GraphemeClass::LVT),
    (0xC7C8, 0xC7C8, GraphemeClass::LV),
    (0xC7C9, 0xC7E3, GraphemeClass::LVT),
    (0xC7E4, 0xC7E4, GraphemeClass::LV),
    (0xC7E5, 0xC7FF, GraphemeClass::LVT),
    (0xC800, 0xC800, GraphemeClass::LV),
    (0xC801, 0xC81B, GraphemeClass::LVT),
    (0xC81C, 0xC81C, GraphemeClass::LV),
    (0xC81D, 0xC837, GraphemeClass::LVT),
    (0xC838, 0xC838, GraphemeClass::LV),
    (0xC839, 0xC853, GraphemeClass::LVT),
    (0xC854, 0xC854, GraphemeClass::LV),
    (0xC855, 0xC86F, GraphemeClass::LVT),
    (0xC870, 0xC870, GraphemeClass::LV),
    (0xC871, 0xC88B, GraphemeClass::LVT),
    (0xC88C, 0xC88C, GraphemeClass::LV),
    (0xC88D, 0xC8A7, GraphemeClass::LVT),
    (0xC8A8, 0xC8A8, GraphemeClass::LV),
    (0xC8A9, 0xC8C3, GraphemeClass::LVT),
    (0xC8C4, 0xC8C4, GraphemeClass::LV),
    (0xC8C5, 0xC8DF, GraphemeClass::LVT),
    (0xC8E0, 0xC8E0, GraphemeClass::LV),
    (0xC8E1, 0xC8FB, GraphemeClass::LVT),
    (0xC8FC, 0xC8FC, GraphemeClass::LV),
    (0xC8FD, 0xC917, GraphemeClass::LVT),
    (0xC918, 0xC918, GraphemeClass::LV),
    (0xC919, 0xC933, GraphemeClass::LVT),
    (0xC934, 0xC934, GraphemeClass::LV),
    (0xC935, 0xC94F, GraphemeClass::LVT),
    (0xC950, 0xC950, GraphemeClass::LV),
    (0xC951, 0xC96B, GraphemeClass::LVT),
    (0xC96C, 0xC96C, GraphemeClass::LV),
    (0xC96D, 0xC987, GraphemeClass::LVT),
    (0xC988, 0xC988, GraphemeClass::LV),
    (0xC989, 0xC9A3, GraphemeClass::LVT),
    (0xC9A4, 0xC9A4, GraphemeClass::LV),
    (0xC9A5, 0xC9BF, GraphemeClass::LVT),
    (0xC9C0, 0xC9C0, GraphemeClass::LV),
    (0xC9C1, 0xC9DB, GraphemeClass::LVT),
    (0xC9DC, 0xC9DC, GraphemeClass::LV),
    (0xC9DD, 0xC9F7, GraphemeClass::LVT),
    (0xC9F8, 0xC9F8, GraphemeClass::LV),
    (0xC9F9, 0xCA13, GraphemeClass::LVT),
    (0xCA14, 0xCA14, GraphemeClass::LV),
    (0xCA15, 0xCA2F, GraphemeClass::LVT),
    (0xCA30, 0xCA30, GraphemeClass::LV),
    (0xCA31, 0xCA4B, GraphemeClass::LVT),
    (0xCA4C, 0xCA4C, GraphemeClass::LV),
    (0xCA4D, 0xCA67, GraphemeClass::LVT),
    (0xCA68, 0xCA68, GraphemeClass::LV),
    (0xCA69, 0xCA83, GraphemeClass::LVT),
    (0xCA84, 0xCA84, GraphemeClass::LV),
    (0xCA85, 0xCA9F, GraphemeClass::LVT),
    (0xCAA0, 0xCAA0, GraphemeClass::LV),
    (0xCAA1, 0xCABB, GraphemeClass::LVT),
    (0xCABC, 0xCABC, GraphemeClass::LV),
    (0xCABD, 0xCAD7, GraphemeClass::LVT),
    (0xCAD8, 0xCAD8, GraphemeClass::LV),
    (0xCAD9, 0xCAF3, GraphemeClass::LVT),
    (0xCAF4, 0xCAF4, GraphemeClass::LV),
    (0xCAF5, 0xCB0F, GraphemeClass::LVT),
    (0xCB10, 0xCB10, GraphemeClass::LV),
    (0xCB11, 0xCB2B, GraphemeClass::LVT),
    (0xCB2C, 0xCB2C, GraphemeClass::LV),
    (0xCB2D, 0xCB47, GraphemeClass::LVT),
    (0xCB48, 0xCB48, GraphemeClass::LV),
    (0xCB49, 0xCB63, GraphemeClass::LVT),
    (0xCB64, 0xCB64, GraphemeClass::LV),
    (0xCB65, 0xCB7F, GraphemeClass::LVT),
    (0xCB80, 0xCB80, GraphemeClass::LV),
    (0xCB81, 0xCB9B, GraphemeClass::LVT),
    (0xCB9C, 0xCB9C, GraphemeClass::LV),
    (0xCB9D, 0xCBB7, GraphemeClass::LVT),
    (0xCBB8, 0xCBB8, GraphemeClass::LV),
    (0xCBB9, 0xCBD3, GraphemeClass::LVT),
    (0xCBD4, 0xCBD4, GraphemeClass::LV),
    (0xCBD5, 0xCBEF, GraphemeClass::LVT),
    (0xCBF0, 0xCBF0, GraphemeClass::LV),
    (0xCBF1, 0xCC0B, GraphemeClass::LVT),
    (0xCC0C, 0xCC0C, GraphemeClass::LV),
    (0xCC0D, 0xCC27, GraphemeClass::LVT),
    (0xCC28, 0xCC28, GraphemeClass::LV),
    (0xCC29, 0xCC43, GraphemeClass::LVT),
    (0xCC44, 0xCC44, GraphemeClass::LV),
    (0xCC45, 0xCC5F, GraphemeClass::LVT),
    (0xCC60, 0xCC60, GraphemeClass::LV),
    (0xCC61, 0xCC7B, GraphemeClass::LVT),
    (0xCC7C, 0xCC7C, GraphemeClass::LV),
    (0xCC7D, 0xCC97, GraphemeClass::LVT),
    (0xCC98, 0xCC98, GraphemeClass::LV),
    (0xCC99, 0xCCB3, GraphemeClass::LVT),
    (0xCCB4, 0xCCB4, GraphemeClass::LV),
    (0xCCB5, 0xCCCF, GraphemeClass::LVT),
    (0xCCD0, 0xCCD0, GraphemeClass::LV),
    (0xCCD1, 0xCCEB, GraphemeClass::LVT),
    (0xCCEC, 0xCCEC, GraphemeClass::LV),
    (0xCCED, 0xCD07, GraphemeClass::LVT),
    (0xCD08, 0xCD08, GraphemeClass::LV),
    (0xCD09, 0xCD23, GraphemeClass::LVT),
    (0xCD24, 0xCD24, GraphemeClass::LV),
    (0xCD25, 0xCD3F, GraphemeClass::LVT),
    (0xCD40, 0xCD40, GraphemeClass::LV),
    (0xCD41, 0xCD5B, GraphemeClass::LVT),
    (0xCD5C, 0xCD5C, GraphemeClass::LV),
    (0xCD5D, 0xCD77, GraphemeClass::LVT),
    (0xCD78, 0xCD78, GraphemeClass::LV),
    (0xCD79, 0xCD93, GraphemeClass::LVT),
    (0xCD94, 0xCD94, GraphemeClass::LV),
    (0xCD95, 0xCDAF, GraphemeClass::LVT),
    (0xCDB0, 0xCDB0, GraphemeClass::LV),
    (0xCDB1, 0xCDCB, GraphemeClass::LVT),
    (0xCDCC, 0xCDCC, GraphemeClass::LV),
    (0xCDCD, 0xCDE7, GraphemeClass::LVT),
    (0xCDE8, 0xCDE8, GraphemeClass::LV),
    (0xCDE9, 0xCE03, GraphemeClass::LVT),
    (0xCE04, 0xCE04, GraphemeClass::LV),
    (0xCE05, 0xCE1F, GraphemeClass::LVT),
    (0xCE20, 0xCE20, GraphemeClass::LV),
    (0xCE21, 0xCE3B, GraphemeClass::LVT),
    (0xCE3C, 0xCE3C, GraphemeClass::LV),
    (0xCE3D, 0xCE57, GraphemeClass::LVT),
    (0xCE58, 0xCE58, GraphemeClass::LV),
    (0xCE59, 0xCE73, GraphemeClass::LVT),
    (0xCE74, 0xCE74, GraphemeClass::LV),
    (0xCE75, 0xCE8F, GraphemeClass::LVT),
    (0xCE90, 0xCE90, GraphemeClass::LV),
    (0xCE91, 0xCEAB, GraphemeClass::LVT),
    (0xCEAC, 0xCEAC, GraphemeClass::LV),
    (0xCEAD, 0xCEC7, GraphemeClass::LVT),
    (0xCEC8, 0xCEC8, GraphemeClass::LV),
    (0xCEC9, 0xCEE3, GraphemeClass::LVT),
    (0xCEE4, 0xCEE4, GraphemeClass::LV),
    (0xCEE5, 0xCEFF, GraphemeClass::LVT),
    (0xCF00, 0xCF00, GraphemeClass::LV),
    (0xCF01, 0xCF1B, GraphemeClass::LVT),
    (0xCF1C, 0xCF1C, GraphemeClass::LV),
    (0xCF1D, 0xCF37, GraphemeClass::LVT),
    (0xCF38, 0xCF38, GraphemeClass::LV),
    (0xCF39, 0xCF53, GraphemeClass::LVT),
    (0xCF54, 0xCF54, GraphemeClass::LV),
    (0xCF55, 0xCF6F, GraphemeClass::LVT),
    (0xCF70, 0xCF70, GraphemeClass::LV),
    (0xCF71, 0xCF8B, GraphemeClass::LVT),
    (0xCF8C, 0xCF8C, GraphemeClass::LV),
    (0xCF8D, 0xCFA7, GraphemeClass::LVT),
    (0xCFA8, 0xCFA8, GraphemeClass::LV),
    (0xCFA9, 0xCFC3, GraphemeClass::LVT),
    (0xCFC4, 0xCFC4, GraphemeClass::LV),
    (0xCFC5, 0xCFDF, GraphemeClass::LVT),
    (0xCFE0, 0xCFE0, GraphemeClass::LV),
    (0xCFE1, 0xCFFB, GraphemeClass::LVT),
    (0xCFFC, 0xCFFC, GraphemeClass::LV),
    (0xCFFD, 0xD017, GraphemeClass::LVT),
    (0xD018, 0xD018, GraphemeClass::LV),
    (0xD019, 0xD033, GraphemeClass::LVT),
    (0xD034, 0xD034, GraphemeClass::LV),
    (0xD035, 0xD04F, GraphemeClass::LVT),
    (0xD050, 0xD050, GraphemeClass::LV),
    (0xD051, 0xD06B, GraphemeClass::LVT),
    (0xD06C, 0xD06C, GraphemeClass::LV),
    (0xD06D, 0xD087, GraphemeClass::LVT),
    (0xD088, 0xD088, GraphemeClass::LV),
    (0xD089, 0xD0A3, GraphemeClass::LVT),
    (0xD0A4, 0xD0A4, GraphemeClass::LV),
    (0xD0A5, 0xD0BF, GraphemeClass::LVT),
    (0xD0C0, 0xD0C0, GraphemeClass::LV),
    (0xD0C1, 0xD0DB, GraphemeClass::LVT),
    (0xD0DC, 0xD0DC, GraphemeClass::LV),
    (0xD0DD, 0xD0F7, GraphemeClass::LVT),
    (0xD0F8, 0xD0F8, GraphemeClass::LV),
    (0xD0F9, 0xD113, GraphemeClass::LVT),
    (0xD114, 0xD114, GraphemeClass::LV),
    (0xD115, 0xD12F, GraphemeClass::LVT),
    (0xD130, 0xD130, GraphemeClass::LV),
    (0xD131, 0xD14B, GraphemeClass::LVT),
    (0xD14C, 0xD14C, GraphemeClass::LV),
    (0xD14D, 0xD167, GraphemeClass::LVT),
    (0xD168, 0xD168, GraphemeClass::LV),
    (0xD169, 0xD183, GraphemeClass::LVT),
    (0xD184, 0xD184, GraphemeClass::LV),
    (0xD185, 0xD19F, GraphemeClass::LVT),
    (0xD1A0, 0xD1A0, GraphemeClass::LV),
    (0xD1A1, 0xD1BB, GraphemeClass::LVT),
    (0xD1BC, 0xD1BC, GraphemeClass::LV),
    (0xD1BD, 0xD1D7, GraphemeClass::LVT),
    (0xD1D8, 0xD1D8, GraphemeClass::LV),
    (0xD1D9, 0xD1F3, GraphemeClass::LVT),
    (0xD1F4, 0xD1F4, GraphemeClass::LV),
    (0xD1F5, 0xD20F, GraphemeClass::LVT),
    (0xD210, 0xD210, GraphemeClass::LV),
    (0xD211, 0xD22B, GraphemeClass::LVT),
    (0xD22C, 0xD22C, GraphemeClass::LV),
    (0xD22D, 0xD247, GraphemeClass::LVT),
    (0xD248, 0xD248, GraphemeClass::LV),
    (0xD249, 0xD263, GraphemeClass::LVT),
    (0xD264, 0xD264, GraphemeClass::LV),
    (0xD265, 0xD27F, GraphemeClass::LVT),
    (0xD280, 0xD280, GraphemeClass::LV),
    (0xD281, 0xD29B, GraphemeClass::LVT),
    (0xD29C, 0xD29C, GraphemeClass::LV),
    (0xD29D, 0xD2B7, GraphemeClass::LVT),
    (0xD2B8, 0xD2B8, GraphemeClass::LV),
    (0xD2B9, 0xD2D3, GraphemeClass::LVT),
    (0xD2D4, 0xD2D4, GraphemeClass::LV),
    (0xD2D5, 0xD2EF, GraphemeClass::LVT),
    (0xD2F0, 0xD2F0, GraphemeClass::LV),
    (0xD2F1, 0xD30B, GraphemeClass::LVT),
    (0xD30C, 0xD30C, GraphemeClass::LV),
    (0xD30D, 0xD327, GraphemeClass::LVT),
    (0xD328, 0xD328, GraphemeClass::LV),
    (0xD329, 0xD343, GraphemeClass::LVT),
    (0xD344, 0xD344, GraphemeClass::LV),
    (0xD345, 0xD35F, GraphemeClass::LVT),
    (0xD360, 0xD360, GraphemeClass::LV),
    (0xD361, 0xD37B, GraphemeClass::LVT),
    (0xD37C, 0xD37C, GraphemeClass::LV),
    (0xD37D, 0xD397, GraphemeClass::LVT),
    (0xD398, 0xD398, GraphemeClass::LV),
    (0xD399, 0xD3B3, GraphemeClass::LVT),
    (0xD3B4, 0xD3B4, GraphemeClass::LV),
    (0xD3B5, 0xD3CF, GraphemeClass::LVT),
    (0xD3D0, 0xD3D0, GraphemeClass::LV),
    (0xD3D1, 0xD3EB, GraphemeClass::LVT),
    (0xD3EC, 0xD3EC, GraphemeClass::LV),
    (0xD3ED, 0xD407, GraphemeClass::LVT),
    (0xD408, 0xD408, GraphemeClass::LV),
    (0xD409, 0xD423, GraphemeClass::LVT),
    (0xD424, 0xD424, GraphemeClass::LV),
    (0xD425, 0xD43F, GraphemeClass::LVT),
    (0xD440, 0xD440, GraphemeClass::LV),
    (0xD441, 0xD45B, GraphemeClass::LVT),
    (0xD45C, 0xD45C, GraphemeClass::LV),
    (0xD45D, 0xD477, GraphemeClass::LVT),
    (0xD478, 0xD478, GraphemeClass::LV),
    (0xD479, 0xD493, GraphemeClass::LVT),
    (0xD494, 0xD494, GraphemeClass::LV),
    (0xD495, 0xD4AF, GraphemeClass::LVT),
    (0xD4B0, 0xD4B0, GraphemeClass::LV),
    (0xD4B1, 0xD4CB, GraphemeClass::LVT),
    (0xD4CC, 0xD4CC, GraphemeClass::LV),
    (0xD4CD, 0xD4E7, GraphemeClass::LVT),
    (0xD4E8, 0xD4E8, GraphemeClass::LV),
    (0xD4E9, 0xD503, GraphemeClass::LVT),
    (0xD504, 0xD504, GraphemeClass::LV),
    (0xD505, 0xD51F, GraphemeClass::LVT),
    (0xD520, 0xD520, GraphemeClass::LV),
    (0xD521, 0xD53B, GraphemeClass::LVT),
    (0xD53C, 0xD53C, GraphemeClass::LV),
    (0xD53D, 0xD557, GraphemeClass::LVT),
    (0xD558, 0xD558, GraphemeClass::LV),
    (0xD559, 0xD573, GraphemeClass::LVT),
    (0xD574, 0xD574, GraphemeClass::LV),
    (0xD575, 0xD58F, GraphemeClass::LVT),
    (0xD590, 0xD590, GraphemeClass::LV),
    (0xD591, 0xD5AB, GraphemeClass::LVT),
    (0xD5AC, 0xD5AC, GraphemeClass::LV),
    (0xD5AD, 0xD5C7, GraphemeClass::LVT),
    (0xD5C8, 0xD5C8, GraphemeClass::LV),
    (0xD5C9, 0xD5E3, GraphemeClass::LVT),
    (0xD5E4, 0xD5E4, GraphemeClass::LV),
    (0xD5E5, 0xD5FF, GraphemeClass::LVT),
    (0xD600, 0xD600, GraphemeClass::LV),
    (0xD601, 0xD61B, GraphemeClass::LVT),
    (0xD61C, 0xD61C, GraphemeClass::LV),
    (0xD61D, 0xD637, GraphemeClass::LVT),
    (0xD638, 0xD638, GraphemeClass::LV),
    (0xD639, 0xD653, GraphemeClass::LVT),
    (0xD654, 0xD654, GraphemeClass::LV),
    (0xD655, 0xD66F, GraphemeClass::LVT),
    (0xD670, 0xD670, GraphemeClass::LV),
    (0xD671, 0xD68B, GraphemeClass::LVT),
    (0xD68C, 0xD68C, GraphemeClass::LV),
    (0xD68D, 0xD6A7, GraphemeClass::LVT),
    (0xD6A8, 0xD6A8, GraphemeClass::LV),
    (0xD6A9, 0xD6C3, GraphemeClass::LVT),
    (0xD6C4, 0xD6C4, GraphemeClass::LV),
    (0xD6C5, 0xD6DF, GraphemeClass::LVT),
    (0xD6E0, 0xD6E0, GraphemeClass::LV),
    (0xD6E1, 0xD6FB, GraphemeClass::LVT),
    (0xD6FC, 0xD6FC, GraphemeClass::LV),
    (0xD6FD, 0xD717, GraphemeClass::LVT),
    (0xD718, 0xD718, GraphemeClass::LV),
    (0xD719, 0xD733, GraphemeClass::LVT),
    (0xD734, 0xD734, GraphemeClass::LV),
    (0xD735, 0xD74F, GraphemeClass::LVT),
    (0xD750, 0xD750, GraphemeClass::LV),
    (0xD751, 0xD76B, GraphemeClass::LVT),
    (0xD76C, 0xD76C, GraphemeClass::LV),
    (0xD76D, 0xD787, GraphemeClass::LVT),
    (0xD788, 0xD788, GraphemeClass::LV),
    (0xD789, 0xD7A3, GraphemeClass::LVT),
    (0xD7B0, 0xD7C6, GraphemeClass::V),
    (0xD7CB, 0xD7FB, GraphemeClass::T),
    (0xFB1E, 0xFB1E, GraphemeClass::ExtendIncb),
    (0xFE00, 0xFE0F, GraphemeClass::ExtendIncb),
    (0xFE20, 0xFE2F, GraphemeClass::ExtendIncb),
    (0xFEFF, 0xFEFF, GraphemeClass::Control),
    (0xFF9E, 0xFF9F, GraphemeClass::ExtendIncb),
    (0xFFF0, 0xFFFB, GraphemeClass::Control),
    (0x101FD, 0x101FD, GraphemeClass::ExtendIncb),
    (0x102E0, 0x102E0, GraphemeClass::ExtendIncb),
    (0x10376, 0x1037A, GraphemeClass::ExtendIncb),
    (0x10A00, 0x10A00, GraphemeClass::Consonant),
    (0x10A01, 0x10A03, GraphemeClass::ExtendIncb),
    (0x10A05, 0x10A06, GraphemeClass::ExtendIncb),
    (0x10A0C, 0x10A0F, GraphemeClass::ExtendIncb),
    (0x10A10, 0x10A13, GraphemeClass::Consonant),
    (0x10A15, 0x10A17, GraphemeClass::Consonant),
    (0x10A19, 0x10A35, GraphemeClass::Consonant),
    (0x10A38, 0x10A3A, GraphemeClass::ExtendIncb),
    (0x10A3F, 0x10A3F, GraphemeClass::Linker),
    (0x10AE5, 0x10AE6, GraphemeClass::ExtendIncb),
    (0x10D24, 0x10D27, GraphemeClass::ExtendIncb),
    (0x10D69, 0x10D6D, GraphemeClass::ExtendIncb),
    (0x10EAB, 0x10EAC, GraphemeClass::ExtendIncb),
    (0x10EFA, 0x10EFF, GraphemeClass::ExtendIncb),
    (0x10F46, 0x10F50, GraphemeClass::ExtendIncb),
    (0x10F82, 0x10F85, GraphemeClass::ExtendIncb),
    (0x11000, 0x11000, GraphemeClass::SpacingMark),
    (0x11001, 0x11001, GraphemeClass::ExtendIncb),
    (0x11002, 0x11002, GraphemeClass::SpacingMark),
    (0x11038, 0x11046, GraphemeClass::ExtendIncb),
    (0x11070, 0x11070, GraphemeClass::ExtendIncb),
    (0x11073, 0x11074, GraphemeClass::ExtendIncb),
    (0x1107F, 0x11081, GraphemeClass::ExtendIncb),
    (0x11082, 0x11082, GraphemeClass::SpacingMark),
    (0x110B0, 0x110B2, GraphemeClass::SpacingMark),
    (0x110B3, 0x110B6, GraphemeClass::ExtendIncb),
    (0x110B7, 0x110B8, GraphemeClass::SpacingMark),
    (0x110B9, 0x110BA, GraphemeClass::ExtendIncb),
    (0x110BD, 0x110BD, GraphemeClass::Prepend),
    (0x110C2, 0x110C2, GraphemeClass::ExtendIncb),
    (0x110CD, 0x110CD, GraphemeClass::Prepend),
    (0x11100, 0x11102, GraphemeClass::ExtendIncb),
    (0x11103, 0x11126, GraphemeClass::Consonant),
    (0x11127, 0x1112B, GraphemeClass::ExtendIncb),
    (0x1112C, 0x1112C, GraphemeClass::SpacingMark),
    (0x1112D, 0x11132, GraphemeClass::ExtendIncb),
    (0x11133, 0x11133, GraphemeClass::Linker),
    (0x11134, 0x11134, GraphemeClass::ExtendIncb),
    (0x11144, 0x11144, GraphemeClass::Consonant),
    (0x11145, 0x11146, GraphemeClass::SpacingMark),
    (0x11147, 0x11147, GraphemeClass::Consonant),
    (0x11173, 0x11173, GraphemeClass::ExtendIncb),
    (0x11180, 0x11181, GraphemeClass::ExtendIncb),
    (0x11182, 0x11182, GraphemeClass::SpacingMark),
    (0x111B3, 0x111B5, GraphemeClass::SpacingMark),
    (0x111B6, 0x111BE, GraphemeClass::ExtendIncb),
    (0x111BF, 0x111BF, GraphemeClass::SpacingMark),
    (0x111C0, 0x111C0, GraphemeClass::ExtendIncb),
    (0x111C2, 0x111C3, GraphemeClass::Prepend),
    (0x111C9, 0x111CC, GraphemeClass::ExtendIncb),
    (0x111CE, 0x111CE, GraphemeClass::SpacingMark),
    (0x111CF, 0x111CF, GraphemeClass::ExtendIncb),
    (0x1122C, 0x1122E, GraphemeClass::SpacingMark),
    (0x1122F, 0x11231, GraphemeClass::ExtendIncb),
    (0x11232, 0x11233, GraphemeClass::SpacingMark),
    (0x11234, 0x11237, GraphemeClass::ExtendIncb),
    (0x1123E, 0x1123E, GraphemeClass::ExtendIncb),
    (0x11241, 0x11241, GraphemeClass::ExtendIncb),
    (0x112DF, 0x112DF, GraphemeClass::ExtendIncb),
    (0x112E0, 0x112E2, GraphemeClass::SpacingMark),
    (0x112E3, 0x112EA, GraphemeClass::ExtendIncb),
    (0x11300, 0x11301, GraphemeClass::ExtendIncb),
    (0x11302, 0x11303, GraphemeClass::SpacingMark),
    (0x1133B, 0x1133C, GraphemeClass::ExtendIncb),
    (0x1133E, 0x1133E, GraphemeClass::ExtendIncb),
    (0x1133F, 0x1133F, GraphemeClass::SpacingMark),
    (0x11340, 0x11340, GraphemeClass::ExtendIncb),
    (0x11341, 0x11344, GraphemeClass::SpacingMark),
    (0x11347, 0x11348, GraphemeClass::SpacingMark),
    (0x1134B, 0x1134C, GraphemeClass::SpacingMark),
    (0x1134D, 0x1134D, GraphemeClass::ExtendIncb),
    (0x11357, 0x11357, GraphemeClass::ExtendIncb),
    (0x11362, 0x11363, GraphemeClass::SpacingMark),
    (0x11366, 0x1136C, GraphemeClass::ExtendIncb),
    (0x11370, 0x11374, GraphemeClass::ExtendIncb),
    (0x11380, 0x11389, GraphemeClass::Consonant),
    (0x1138B, 0x1138B, GraphemeClass::Consonant),
    (0x1138E, 0x1138E, GraphemeClass::Consonant),
    (0x11390, 0x113B5, GraphemeClass::Consonant),
    (0x113B8, 0x113B8, GraphemeClass::ExtendIncb),
    (0x113B9, 0x113BA, GraphemeClass::SpacingMark),
    (0x113BB, 0x113C0, GraphemeClass::ExtendIncb),
    (0x113C2, 0x113C2, GraphemeClass::ExtendIncb),
    (0x113C5, 0x113C5, GraphemeClass::ExtendIncb),
    (0x113C7, 0x113C9, GraphemeClass::ExtendIncb),
    (0x113CA, 0x113CA, GraphemeClass::SpacingMark),
    (0x113CC, 0x113CD, GraphemeClass::SpacingMark),
    (0x113CE, 0x113CF, GraphemeClass::ExtendIncb),
    (0x113D0, 0x113D0, GraphemeClass::Linker),
    (0x113D1, 0x113D1, GraphemeClass::Prepend),
    (0x113D2, 0x113D2, GraphemeClass::ExtendIncb),
    (0x113E1, 0x113E2, GraphemeClass::ExtendIncb),
    (0x11435, 0x11437, GraphemeClass::SpacingMark),
    (0x11438, 0x1143F, GraphemeClass::ExtendIncb),
    (0x11440, 0x11441, GraphemeClass::SpacingMark),
    (0x11442, 0x11444, GraphemeClass::ExtendIncb),
    (0x11445, 0x11445, GraphemeClass::SpacingMark),
    (0x11446, 0x11446, GraphemeClass::ExtendIncb),
    (0x1145E, 0x1145E, GraphemeClass::ExtendIncb),
    (0x114B0, 0x114B0, GraphemeClass::ExtendIncb),
    (0x114B1, 0x114B2, GraphemeClass::SpacingMark),
    (0x114B3, 0x114B8, GraphemeClass::ExtendIncb),
    (0x114B9, 0x114B9, GraphemeClass::SpacingMark),
    (0x114BA, 0x114BA, GraphemeClass::ExtendIncb),
    (0x114BB, 0x114BC, GraphemeClass::SpacingMark),
    (0x114BD, 0x114BD, GraphemeClass::ExtendIncb),
    (0x114BE, 0x114BE, GraphemeClass::SpacingMark),
    (0x114BF, 0x114C0, GraphemeClass::ExtendIncb),
    (0x114C1, 0x114C1, GraphemeClass::SpacingMark),
    (0x114C2, 0x114C3, GraphemeClass::ExtendIncb),
    (0x115AF, 0x115AF, GraphemeClass::ExtendIncb),
    (0x115B0, 0x115B1, GraphemeClass::SpacingMark),
    (0x115B2, 0x115B5, GraphemeClass::ExtendIncb),
    (0x115B8, 0x115BB, GraphemeClass::SpacingMark),
    (0x115BC, 0x115BD, GraphemeClass::ExtendIncb),
    (0x115BE, 0x115BE, GraphemeClass::SpacingMark),
    (0x115BF, 0x115C0, GraphemeClass::ExtendIncb),
    (0x115DC, 0x115DD, GraphemeClass::ExtendIncb),
    (0x11630, 0x11632, GraphemeClass::SpacingMark),
    (0x11633, 0x1163A, GraphemeClass::ExtendIncb),
    (0x1163B, 0x1163C, GraphemeClass::SpacingMark),
    (0x1163D, 0x1163D, GraphemeClass::ExtendIncb),
    (0x1163E, 0x1163E, GraphemeClass::SpacingMark),
    (0x1163F, 0x11640, GraphemeClass::ExtendIncb),
    (0x116AB, 0x116AB, GraphemeClass::ExtendIncb),
    (0x116AC, 0x116AC, GraphemeClass::SpacingMark),
    (0x116AD, 0x116AD, GraphemeClass::ExtendIncb),
    (0x116AE, 0x116AF, GraphemeClass::SpacingMark),
    (0x116B0, 0x116B7, GraphemeClass::ExtendIncb),
    (0x1171D, 0x1171D, GraphemeClass::ExtendIncb),
    (0x1171E, 0x1171E, GraphemeClass::SpacingMark),
    (0x1171F, 0x1171F, GraphemeClass::ExtendIncb),
    (0x11722, 0x11725, GraphemeClass::ExtendIncb),
    (0x11726, 0x11726, GraphemeClass::SpacingMark),
    (0x11727, 0x1172B, GraphemeClass::ExtendIncb),
    (0x1182C, 0x1182E, GraphemeClass::SpacingMark),
    (0x1182F, 0x11837, GraphemeClass::ExtendIncb),
    (0x11838, 0x11838, GraphemeClass::SpacingMark),
    (0x11839, 0x1183A, GraphemeClass::ExtendIncb),
    (0x11900, 0x11906, GraphemeClass::Consonant),
    (0x11909, 0x11909, GraphemeClass::Consonant),
    (0x1190C, 0x11913, GraphemeClass::Consonant),
    (0x11915, 0x11916, GraphemeClass::Consonant),
    (0x11918, 0x1192F, GraphemeClass::Consonant),
    (0x11930, 0x11930, GraphemeClass::ExtendIncb),
    (0x11931, 0x11935, GraphemeClass::SpacingMark),
    (0x11937, 0x11938, GraphemeClass::SpacingMark),
    (0x1193B, 0x1193D, GraphemeClass::ExtendIncb),
    (0x1193E, 0x1193E, GraphemeClass::Linker),
    (0x1193F, 0x1193F, GraphemeClass::Prepend),
    (0x11940, 0x11940, GraphemeClass::SpacingMark),
    (0x11941, 0x11941, GraphemeClass::Prepend),
    (0x11942, 0x11942, GraphemeClass::SpacingMark),
    (0x11943, 0x11943, GraphemeClass::ExtendIncb),
    (0x119D1, 0x119D3, GraphemeClass::SpacingMark),
    (0x119D4, 0x119D7, GraphemeClass::ExtendIncb),
    (0x119DA, 0x119DB, GraphemeClass::ExtendIncb),
    (0x119DC, 0x119DF, GraphemeClass::SpacingMark),
    (0x119E0, 0x119E0, GraphemeClass::ExtendIncb),
    (0x119E4, 0x119E4, GraphemeClass::SpacingMark),
    (0x11A00, 0x11A00, GraphemeClass::Consonant),
    (0x11A01, 0x11A0A, GraphemeClass::ExtendIncb),
    (0x11A0B, 0x11A32, GraphemeClass::Consonant),
    (0x11A33, 0x11A38, GraphemeClass::ExtendIncb),
    (0x11A39, 0x11A39, GraphemeClass::SpacingMark),
    (0x11A3B, 0x11A3E, GraphemeClass::ExtendIncb),
    (0x11A47, 0x11A47, GraphemeClass::Linker),
    (0x11A50, 0x11A50, GraphemeClass::Consonant),
    (0x11A51, 0x11A56, GraphemeClass::ExtendIncb),
    (0x11A57, 0x11A58, GraphemeClass::SpacingMark),
    (0x11A59, 0x11A5B, GraphemeClass::ExtendIncb),
    (0x11A5C, 0x11A83, GraphemeClass::Consonant),
    (0x11A84, 0x11A89, GraphemeClass::Prepend),
    (0x11A8A, 0x11A96, GraphemeClass::ExtendIncb),
    (0x11A97, 0x11A97, GraphemeClass::SpacingMark),
    (0x11A98, 0x11A98, GraphemeClass::ExtendIncb),
    (0x11A99, 0x11A99, GraphemeClass::Linker),
    (0x11B60, 0x11B60, GraphemeClass::ExtendIncb),
    (0x11B61, 0x11B61, GraphemeClass::SpacingMark),
    (0x11B62, 0x11B64, GraphemeClass::ExtendIncb),
    (0x11B65, 0x11B65, GraphemeClass::SpacingMark),
    (0x11B66, 0x11B66, GraphemeClass::ExtendIncb),
    (0x11B67, 0x11B67, GraphemeClass::SpacingMark),
    (0x11C2F, 0x11C2F, GraphemeClass::SpacingMark),
    (0x11C30, 0x11C36, GraphemeClass::ExtendIncb),
    (0x11C38, 0x11C3D, GraphemeClass::ExtendIncb),
    (0x11C3E, 0x11C3E, GraphemeClass::SpacingMark),
    (0x11C3F, 0x11C3F, GraphemeClass::ExtendIncb),
    (0x11C92, 0x11CA7, GraphemeClass::ExtendIncb),
    (0x11CA9, 0x11CA9, GraphemeClass::SpacingMark),
    (0x11CAA, 0x11CB0, GraphemeClass::ExtendIncb),
    (0x11CB1, 0x11CB1, GraphemeClass::SpacingMark),
    (0x11CB2, 0x11CB3, GraphemeClass::ExtendIncb),
    (0x11CB4, 0x11CB4, GraphemeClass::SpacingMark),
    (0x11CB5, 0x11CB6, GraphemeClass::ExtendIncb),
    (0x11D31, 0x11D36, GraphemeClass::ExtendIncb),
    (0x11D3A, 0x11D3A, GraphemeClass::ExtendIncb),
    (0x11D3C, 0x11D3D, GraphemeClass::ExtendIncb),
    (0x11D3F, 0x11D45, GraphemeClass::ExtendIncb),
    (0x11D46, 0x11D46, GraphemeClass::Prepend),
    (0x11D47, 0x11D47, GraphemeClass::ExtendIncb),
    (0x11D8A, 0x11D8E, GraphemeClass::SpacingMark),
    (0x11D90, 0x11D91, GraphemeClass::ExtendIncb),
    (0x11D93, 0x11D94, GraphemeClass::SpacingMark),
    (0x11D95, 0x11D95, GraphemeClass::ExtendIncb),
    (0x11D96, 0x11D96, GraphemeClass::SpacingMark),
    (0x11D97, 0x11D97, GraphemeClass::ExtendIncb),
    (0x11EF3, 0x11EF4, GraphemeClass::ExtendIncb),
    (0x11EF5, 0x11EF6, GraphemeClass::SpacingMark),
    (0x11F00, 0x11F01, GraphemeClass::ExtendIncb),
    (0x11F02, 0x11F02, GraphemeClass::Prepend),
    (0x11F03, 0x11F03, GraphemeClass::SpacingMark),
    (0x11F04, 0x11F10, GraphemeClass::Consonant),
    (0x11F12, 0x11F33, GraphemeClass::Consonant),
    (0x11F34, 0x11F35, GraphemeClass::SpacingMark),
    (0x11F36, 0x11F3A, GraphemeClass::ExtendIncb),
    (0x11F3E, 0x11F3F, GraphemeClass::SpacingMark),
    (0x11F40, 0x11F41, GraphemeClass::ExtendIncb),
    (0x11F42, 0x11F42, GraphemeClass::Linker),
    (0x11F5A, 0x11F5A, GraphemeClass::ExtendIncb),
    (0x13430, 0x1343F, GraphemeClass::Control),
    (0x13440, 0x13440, GraphemeClass::ExtendIncb),
    (0x13447, 0x13455, GraphemeClass::ExtendIncb),
    (0x1611E, 0x16129, GraphemeClass::ExtendIncb),
    (0x1612A, 0x1612C, GraphemeClass::SpacingMark),
    (0x1612D, 0x1612F, GraphemeClass::ExtendIncb),
    (0x16AF0, 0x16AF4, GraphemeClass::ExtendIncb),
    (0x16B30, 0x16B36, GraphemeClass::ExtendIncb),
    (0x16D63, 0x16D63, GraphemeClass::V),
    (0x16D67, 0x16D6A, GraphemeClass::V),
    (0x16F4F, 0x16F4F, GraphemeClass::ExtendIncb),
    (0x16F51, 0x16F87, GraphemeClass::SpacingMark),
    (0x16F8F, 0x16F92, GraphemeClass::ExtendIncb),
    (0x16FE4, 0x16FE4, GraphemeClass::ExtendIncb),
    (0x16FF0, 0x16FF1, GraphemeClass::ExtendIncb),
    (0x1BC9D, 0x1BC9E, GraphemeClass::ExtendIncb),
    (0x1BCA0, 0x1BCA3, GraphemeClass::Control),
    (0x1CF00, 0x1CF2D, GraphemeClass::ExtendIncb),
    (0x1CF30, 0x1CF46, GraphemeClass::ExtendIncb),
    (0x1D165, 0x1D169, GraphemeClass::ExtendIncb),
    (0x1D16D, 0x1D172, GraphemeClass::ExtendIncb),
    (0x1D173, 0x1D17A, GraphemeClass::Control),
    (0x1D17B, 0x1D182, GraphemeClass::ExtendIncb),
    (0x1D185, 0x1D18B, GraphemeClass::ExtendIncb),
    (0x1D1AA, 0x1D1AD, GraphemeClass::ExtendIncb),
    (0x1D242, 0x1D244, GraphemeClass::ExtendIncb),
    (0x1DA00, 0x1DA36, GraphemeClass::ExtendIncb),
    (0x1DA3B, 0x1DA6C, GraphemeClass::ExtendIncb),
    (0x1DA75, 0x1DA75, GraphemeClass::ExtendIncb),
    (0x1DA84, 0x1DA84, GraphemeClass::ExtendIncb),
    (0x1DA9B, 0x1DA9F, GraphemeClass::ExtendIncb),
    (0x1DAA1, 0x1DAAF, GraphemeClass::ExtendIncb),
    (0x1E000, 0x1E006, GraphemeClass::ExtendIncb),
    (0x1E008, 0x1E018, GraphemeClass::ExtendIncb),
    (0x1E01B, 0x1E021, GraphemeClass::ExtendIncb),
    (0x1E023, 0x1E024, GraphemeClass::ExtendIncb),
    (0x1E026, 0x1E02A, GraphemeClass::ExtendIncb),
    (0x1E08F, 0x1E08F, GraphemeClass::ExtendIncb),
    (0x1E130, 0x1E136, GraphemeClass::ExtendIncb),
    (0x1E2AE, 0x1E2AE, GraphemeClass::ExtendIncb),
    (0x1E2EC, 0x1E2EF, GraphemeClass::ExtendIncb),
    (0x1E4EC, 0x1E4EF, GraphemeClass::ExtendIncb),
    (0x1E5EE, 0x1E5EF, GraphemeClass::ExtendIncb),
    (0x1E6E3, 0x1E6E3, GraphemeClass::ExtendIncb),
    (0x1E6E6, 0x1E6E6, GraphemeClass::ExtendIncb),
    (0x1E6EE, 0x1E6EF, GraphemeClass::ExtendIncb),
    (0x1E6F5, 0x1E6F5, GraphemeClass::ExtendIncb),
    (0x1E8D0, 0x1E8D6, GraphemeClass::ExtendIncb),
    (0x1E944, 0x1E94A, GraphemeClass::ExtendIncb),
    (0x1F004, 0x1F004, GraphemeClass::ExtPic),
    (0x1F02C, 0x1F02F, GraphemeClass::ExtPic),
    (0x1F094, 0x1F09F, GraphemeClass::ExtPic),
    (0x1F0AF, 0x1F0B0, GraphemeClass::ExtPic),
    (0x1F0C0, 0x1F0C0, GraphemeClass::ExtPic),
    (0x1F0CF, 0x1F0D0, GraphemeClass::ExtPic),
    (0x1F0F6, 0x1F0FF, GraphemeClass::ExtPic),
    (0x1F170, 0x1F171, GraphemeClass::ExtPic),
    (0x1F17E, 0x1F17F, GraphemeClass::ExtPic),
    (0x1F18E, 0x1F18E, GraphemeClass::ExtPic),
    (0x1F191, 0x1F19A, GraphemeClass::ExtPic),
    (0x1F1AE, 0x1F1E5, GraphemeClass::ExtPic),
    (0x1F1E6, 0x1F1FF, GraphemeClass::RegionalIndicator),
    (0x1F201, 0x1F20F, GraphemeClass::ExtPic),
    (0x1F21A, 0x1F21A, GraphemeClass::ExtPic),
    (0x1F22F, 0x1F22F, GraphemeClass::ExtPic),
    (0x1F232, 0x1F23A, GraphemeClass::ExtPic),
    (0x1F23C, 0x1F23F, GraphemeClass::ExtPic),
    (0x1F249, 0x1F25F, GraphemeClass::ExtPic),
    (0x1F266, 0x1F321, GraphemeClass::ExtPic),
    (0x1F324, 0x1F393, GraphemeClass::ExtPic),
    (0x1F396, 0x1F397, GraphemeClass::ExtPic),
    (0x1F399, 0x1F39B, GraphemeClass::ExtPic),
    (0x1F39E, 0x1F3F0, GraphemeClass::ExtPic),
    (0x1F3F3, 0x1F3F5, GraphemeClass::ExtPic),
    (0x1F3F7, 0x1F3FA, GraphemeClass::ExtPic),
    (0x1F3FB, 0x1F3FF, GraphemeClass::ExtendIncb),
    (0x1F400, 0x1F4FD, GraphemeClass::ExtPic),
    (0x1F4FF, 0x1F53D, GraphemeClass::ExtPic),
    (0x1F549, 0x1F54E, GraphemeClass::ExtPic),
    (0x1F550, 0x1F567, GraphemeClass::ExtPic),
    (0x1F56F, 0x1F570, GraphemeClass::ExtPic),
    (0x1F573, 0x1F57A, GraphemeClass::ExtPic),
    (0x1F587, 0x1F587, GraphemeClass::ExtPic),
    (0x1F58A, 0x1F58D, GraphemeClass::ExtPic),
    (0x1F590, 0x1F590, GraphemeClass::ExtPic),
    (0x1F595, 0x1F596, GraphemeClass::ExtPic),
    (0x1F5A4, 0x1F5A5, GraphemeClass::ExtPic),
    (0x1F5A8, 0x1F5A8, GraphemeClass::ExtPic),
    (0x1F5B1, 0x1F5B2, GraphemeClass::ExtPic),
    (0x1F5BC, 0x1F5BC, GraphemeClass::ExtPic),
    (0x1F5C2, 0x1F5C4, GraphemeClass::ExtPic),
    (0x1F5D1, 0x1F5D3, GraphemeClass::ExtPic),
    (0x1F5DC, 0x1F5DE, GraphemeClass::ExtPic),
    (0x1F5E1, 0x1F5E1, GraphemeClass::ExtPic),
    (0x1F5E3, 0x1F5E3, GraphemeClass::ExtPic),
    (0x1F5E8, 0x1F5E8, GraphemeClass::ExtPic),
    (0x1F5EF, 0x1F5EF, GraphemeClass::ExtPic),
    (0x1F5F3, 0x1F5F3, GraphemeClass::ExtPic),
    (0x1F5FA, 0x1F64F, GraphemeClass::ExtPic),
    (0x1F680, 0x1F6C5, GraphemeClass::ExtPic),
    (0x1F6CB, 0x1F6D2, GraphemeClass::ExtPic),
    (0x1F6D5, 0x1F6E5, GraphemeClass::ExtPic),
    (0x1F6E9, 0x1F6E9, GraphemeClass::ExtPic),
    (0x1F6EB, 0x1F6F0, GraphemeClass::ExtPic),
    (0x1F6F3, 0x1F6FF, GraphemeClass::ExtPic),
    (0x1F7DA, 0x1F7FF, GraphemeClass::ExtPic),
    (0x1F80C, 0x1F80F, GraphemeClass::ExtPic),
    (0x1F848, 0x1F84F, GraphemeClass::ExtPic),
    (0x1F85A, 0x1F85F, GraphemeClass::ExtPic),
    (0x1F888, 0x1F88F, GraphemeClass::ExtPic),
    (0x1F8AE, 0x1F8AF, GraphemeClass::ExtPic),
    (0x1F8BC, 0x1F8BF, GraphemeClass::ExtPic),
    (0x1F8C2, 0x1F8CF, GraphemeClass::ExtPic),
    (0x1F8D9, 0x1F8FF, GraphemeClass::ExtPic),
    (0x1F90C, 0x1F93A, GraphemeClass::ExtPic),
    (0x1F93C, 0x1F945, GraphemeClass::ExtPic),
    (0x1F947, 0x1F9FF, GraphemeClass::ExtPic),
    (0x1FA58, 0x1FA5F, GraphemeClass::ExtPic),
    (0x1FA6E, 0x1FAFF, GraphemeClass::ExtPic),
    (0x1FC00, 0x1FFFD, GraphemeClass::ExtPic),
    (0xE0000, 0xE001F, GraphemeClass::Control),
    (0xE0020, 0xE007F, GraphemeClass::ExtendIncb),
    (0xE0080, 0xE00FF, GraphemeClass::Control),
    (0xE0100, 0xE01EF, GraphemeClass::ExtendIncb),
    (0xE01F0, 0xE0FFF, GraphemeClass::Control),
];
